/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

/* private headers */
#include "pfmlib_priv.h"			/* library private */
#include "pfmlib_e2k_priv.h"

const pfmlib_attr_desc_t e2k_mods[]={
	PFM_ATTR_B("k", "monitor at priv level 0"),		/* monitor priv level 0 */
	PFM_ATTR_B("u", "monitor at priv level 1, 2, 3"),	/* monitor priv level 1, 2, 3 */
	PFM_ATTR_NULL /* end-marker to avoid exporting number of entries */
};
#define E2K_NUM_MODS (sizeof(e2k_mods)/sizeof(pfmlib_attr_desc_t) - 1)

int
pfm_e2k_detect(void *this)
{
	pfmlib_pmu_t *pmu = this;
	pfm_pmu_t model;
	if(__builtin_cpu_is ("elbrus-v1"))
		model = PFM_PMU_E2K_V1;
	else if(__builtin_cpu_is ("elbrus-v2"))
		model = PFM_PMU_E2K_V2;
	else if(__builtin_cpu_is ("elbrus-v3"))
		model = PFM_PMU_E2K_V3;
	else if(__builtin_cpu_is ("elbrus-v4"))
	        model = PFM_PMU_E2K_V4;
	else if(__builtin_cpu_is ("elbrus-v5"))
		model = PFM_PMU_E2K_V5;
	else if(__builtin_cpu_is ("elbrus-v6"))
                model = PFM_PMU_E2K_V6;
	else
		return PFM_ERR_NOTSUPP;
	return model == pmu->pmu ? PFM_SUCCESS : PFM_ERR_NOTSUPP;
}

void
pfm_e2k_display_reg(void *this, pfmlib_event_desc_t *e, pfm_e2k_reg_t reg)
{
	__pfm_vbprintf("[0x%x code=0x%x ctrl_s0=%d ctrl_s1=%d ctrl_s2=%d ctrl_s3=%d ] %s\n",
		reg.val,
		reg.config.code,
		reg.config.ctrl_s0,
		reg.config.ctrl_s1,
		reg.config.ctrl_s2,
		reg.config.ctrl_s3,
		e->fstr);
}

int
pfm_e2k_get_encoding(void *this, pfmlib_event_desc_t *e)
{
	const e2k_entry_t *pe = this_pe(this);
	pfm_event_attr_info_t *a;
	pfm_e2k_reg_t reg;
	int i;

	reg.val = pe[e->event].code;

	for (i = 0; i < e->nattrs; i++) {
		a = attr(e, i);
		if (a->ctrl != PFM_ATTR_CTRL_PMU)
			continue;
	}

	e->count = 2;
	e->codes[0] = reg.val;
	if (pe[e->event].mask == 0x1)
		e->codes[1] = 0x0;
	else if (pe[e->event].mask == 0x2)
	        e->codes[1] = 0x1;
	else if (pe[e->event].mask == 0x4)
		e->codes[1] = 0x2;
	else if (pe[e->event].mask == 0xc)
	{	
		if (pe[e->event].code % 2 ==0)
			e->codes[1] = 0x2; 
		else
			e->codes[1] = 0x3;

		if (pe[e->event].code == 0x14) //in case of EXEC event
		        e->codes[1] = 0x3;

		if (pe[e->event].code == 0x01) //in case of IB_AHEAD_MISS event
			e->codes[1] = 0x2;

	}
	evt_strcat(e->fstr, "%s", pe[e->event].name);

	pfmlib_sort_attr(e);
	for (i = 0; i < e->nattrs; i++) {
		a = attr(e, i);
		if (a->ctrl != PFM_ATTR_CTRL_PMU)
			continue;
	}

	pfm_e2k_display_reg(this, e, reg);
	return PFM_SUCCESS;
}

int
pfm_e2k_get_event_first(void *this)
{
	return 0;
}

int
pfm_e2k_get_event_next(void *this, int idx)
{
	pfmlib_pmu_t *p = this;
	if (idx >= (p->pme_count-1))
		return -1;
	return idx+1;
}

int
pfm_e2k_event_is_valid(void *this, int pidx)
{
	pfmlib_pmu_t *p = this;
	return pidx >= 0 && pidx < p->pme_count;
}

int
pfm_e2k_validate_table(void *this, FILE *fp)
{

	pfmlib_pmu_t *pmu = this;
	const e2k_entry_t *pe = this_pe(this);
	int i, j, error = 0;

	for(i=0; i < pmu->pme_count; i++) {
		if (!pe[i].name) {
			fprintf(fp, "pmu: %s event%d: :: no name (prev event was %s)\n", pmu->name, i,
			i > 1 ? pe[i-1].name : "??");
			error++;
		}
		if (!pe[i].desc) {
			fprintf(fp, "pmu: %s event%d: %s :: no description\n", pmu->name, i, pe[i].name);
			error++;
		}
		for(j=i+1; j < pmu->pme_count; j++) {
			if (pe[i].code == pe[j].code && pe[i].mask == pe[j].mask) {
				fprintf(fp, "pmu: %s event%d: %s code: 0x%x mask: 0x%x is duplicated in event%d : %s\n", pmu->name, i, pe[i].name, pe[i].code, j, pe[i].mask, pe[j].name);
				error++;
			}
		}
	}
	return error ? PFM_ERR_INVAL : PFM_SUCCESS;
}

int
pfm_e2k_get_event_attr_info(void *this, int pidx, int attr_idx, pfm_event_attr_info_t *info)
{
	info->name = e2k_mods[attr_idx].name;
	info->desc = e2k_mods[attr_idx].desc;
	info->type = e2k_mods[attr_idx].type;
	info->equiv= NULL;
	info->code = attr_idx;
	info->is_dfl = 0;
	info->is_precise = 0;
	info->ctrl = PFM_ATTR_CTRL_PMU;;

	return PFM_SUCCESS;
}

int
pfm_e2k_get_event_info(void *this, int idx, pfm_event_info_t *info)
{
	pfmlib_pmu_t *pmu = this;
	const e2k_entry_t *pe = this_pe(this);
	info->name  = pe[idx].name;
	info->desc  = pe[idx].desc;
	info->code  = pe[idx].code;
	info->equiv = NULL;
	info->idx   = idx;
	info->pmu   = pmu->pmu;
	info->is_precise = 0;

	info->nattrs  = pfm_e2k_get_event_nattrs(this, idx);

	return PFM_SUCCESS;
}

unsigned int
pfm_e2k_get_event_nattrs(void *this, int pidx)
{

	return E2K_NUM_ATTRS;
}
