/* Copyright (c) 2020-2024 AO MCST
 * Distributed under the terms of MIT License.
 */

#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

/* private headers */
#include "pfmlib_priv.h"			/* library private */
#include "pfmlib_mcst_sparc_priv.h"

const pfmlib_attr_desc_t mcst_sparc_mods[]={
	PFM_ATTR_B("k", "monitor at priv level 0"),		/* monitor priv level 0 */
	PFM_ATTR_B("u", "monitor at priv level 1, 2, 3"),	/* monitor priv level 1, 2, 3 */
	PFM_ATTR_NULL /* end-marker to avoid exporting number of entries */
};
#define MCST_SPARC_NUM_MODS (sizeof(mcst_sparc_mods)/sizeof(pfmlib_attr_desc_t) - 1)

int
pfm_mcst_sparc_detect(void *this)
{
	pfmlib_pmu_t *pmu = this;
	pfm_pmu_t model;
	#ifdef __r1000__
		model = PFM_PMU_MCST_SPARC_R1000;
	#else
		return PFM_ERR_NOTSUPP;
	#endif
	return model == pmu->pmu ? PFM_SUCCESS : PFM_ERR_NOTSUPP;
}

void
pfm_mcst_sparc_display_reg(void *this, pfmlib_event_desc_t *e, pfm_mcst_sparc_reg_t reg)
{
	__pfm_vbprintf("[0x%x code=0x%x ctrl_s0=%d ctrl_s1=%d ctrl_s2=%d ctrl_s3=%d ] %s\n",
		reg.val,
		reg.config.code,
		reg.config.ctrl_s0,
		reg.config.ctrl_s1,
		reg.config.ctrl_s2,
		reg.config.ctrl_s3,
		e->fstr);
}

int
pfm_mcst_sparc_get_encoding(void *this, pfmlib_event_desc_t *e)
{
	const mcst_sparc_entry_t *pe = this_pe(this);
	pfm_event_attr_info_t *a;
	pfm_mcst_sparc_reg_t reg;
	int i;

	reg.val = pe[e->event].code;

	for (i = 0; i < e->nattrs; i++) {
		a = attr(e, i);
		if (a->ctrl != PFM_ATTR_CTRL_PMU)
			continue;
	}

	e->count = 2;
	e->codes[0] = reg.val;
	e->codes[1] = pe[e->event].mask;

	evt_strcat(e->fstr, "%s", pe[e->event].name);

	pfmlib_sort_attr(e);
	for (i = 0; i < e->nattrs; i++) {
		a = attr(e, i);
		if (a->ctrl != PFM_ATTR_CTRL_PMU)
			continue;
	}

	pfm_mcst_sparc_display_reg(this, e, reg);
	return PFM_SUCCESS;
}

int
pfm_mcst_sparc_get_event_first(void *this)
{
	return 0;
}

int
pfm_mcst_sparc_get_event_next(void *this, int idx)
{
	pfmlib_pmu_t *p = this;
	if (idx >= (p->pme_count-1))
		return -1;
	return idx+1;
}

int
pfm_mcst_sparc_event_is_valid(void *this, int pidx)
{
	pfmlib_pmu_t *p = this;
	return pidx >= 0 && pidx < p->pme_count;
}

int
pfm_mcst_sparc_validate_table(void *this, FILE *fp)
{

	pfmlib_pmu_t *pmu = this;
	const mcst_sparc_entry_t *pe = this_pe(this);
	int i, j, error = 0;

	for(i=0; i < pmu->pme_count; i++) {
		if (!pe[i].name) {
			fprintf(fp, "pmu: %s event%d: :: no name (prev event was %s)\n", pmu->name, i,
			i > 1 ? pe[i-1].name : "??");
			error++;
		}
		if (!pe[i].desc) {
			fprintf(fp, "pmu: %s event%d: %s :: no description\n", pmu->name, i, pe[i].name);
			error++;
		}
		for(j=i+1; j < pmu->pme_count; j++) {
			if (pe[i].code == pe[j].code && pe[i].mask == pe[j].mask) {
				fprintf(fp, "pmu: %s event%d: %s code: 0x%x mask: 0x%x is duplicated in event%d : %s\n", pmu->name, i, pe[i].name, pe[i].code, j, pe[i].mask, pe[j].name);
				error++;
			}
		}
	}
	return error ? PFM_ERR_INVAL : PFM_SUCCESS;
}

int
pfm_mcst_sparc_get_event_attr_info(void *this, int pidx, int attr_idx, pfm_event_attr_info_t *info)
{
	info->name = mcst_sparc_mods[attr_idx].name;
	info->desc = mcst_sparc_mods[attr_idx].desc;
	info->type = mcst_sparc_mods[attr_idx].type;
	info->equiv= NULL;
	info->code = attr_idx;
	info->is_dfl = 0;
	info->is_precise = 0;
	info->ctrl = PFM_ATTR_CTRL_PMU;;

	return PFM_SUCCESS;
}

int
pfm_mcst_sparc_get_event_info(void *this, int idx, pfm_event_info_t *info)
{
	pfmlib_pmu_t *pmu = this;
	const mcst_sparc_entry_t *pe = this_pe(this);
	info->name  = pe[idx].name;
	info->desc  = pe[idx].desc;
	info->code  = pe[idx].code;
	info->equiv = NULL;
	info->idx   = idx;
	info->pmu   = pmu->pmu;
	info->is_precise = 0;

	info->nattrs  = pfm_mcst_sparc_get_event_nattrs(this, idx);

	return PFM_SUCCESS;
}

unsigned int
pfm_mcst_sparc_get_event_nattrs(void *this, int pidx)
{

	return MCST_SPARC_NUM_ATTRS;
}
