#include "elo.h"
#include "elo_usb.h"
#include "elo_rs232.h"

/**
 * Filter for the AutoDevProbe scandir on /dev/input.
 *
 * @param dir The current directory entry provided by scandir.
 *
 * @return Non-zero if the given directory entry starts with "event", or zero
 * otherwise.
 */
static int
EloIsEventDevice(const struct dirent *dir)
{
    return strncmp(ELO_EVENT_DEV_NAME, dir->d_name, 5) == 0;
}

static Bool
EloDeviceIsFree(char *fileName)
{
    InputInfoPtr pInput = xf86FirstLocalDevice();
    Bool isFree = TRUE;

    while (pInput) {
        if (pInput->type_name == XI_TOUCHSCREEN) {
            EloDevicePtr pPriv = (EloDevicePtr) (pInput->private);
            if (pPriv && pPriv->device) {
                if (strcmp(fileName, pPriv->device) == 0) {
                    isFree = FALSE;
                    break;
                }
            }
        }
        pInput = pInput->next;
    }
    return isFree;
}

static Bool
EloGetDevice(InputInfoPtr pInfo)
{
    EloDevicePtr pPriv;
    struct dirent **nameList;
    char *pStr;
    int count;
    int i;

    if (!pInfo)
        return FALSE;

    pPriv = (EloDevicePtr) pInfo->private;
    if (!pPriv)
        return FALSE;

    /* Try to get "Device" option value from configuration file */
    pStr = xf86SetStrOption(pInfo->options, "Device", NULL);
    if (pStr) {
        if (EloDeviceIsFree(pStr)) {
            pPriv->device = strdup(pStr);
            return TRUE;
        }
        else {
            xf86Msg(X_ERROR, "%s: Device %s is already used \n", pInfo->name,
                    pStr);
            return FALSE;
        }
    }

    count = scandir(ELO_DEV_INPUT_EVENT, &nameList, EloIsEventDevice, 0);
    if (count <= 0)
        return FALSE;

    for (i = 0; i < count; i++) {
        char deviceName[256] = "???";
        char fileName[64];
        int ret;
        int fd;

        snprintf(fileName, sizeof(fileName), "%s/%s",
                 ELO_DEV_INPUT_EVENT, nameList[i]->d_name);

        fd = open(fileName, O_RDONLY);
        if (fd < 0)
            continue;

        ret = ioctl(fd, EVIOCGNAME(sizeof(deviceName)), deviceName);
        close(fd);

        if (ret < 0)
            continue;

        if (strstr(deviceName, ELO_DEVICE_NAME) != NULL)
            if (EloDeviceIsFree(fileName)) {
                pPriv->device = strdup(fileName);
                break;
            }
    }

    for (i = 0; i < count; i++) {
        if (nameList[i])
            free(nameList[i]);
    }

    if (pPriv->device == NULL) {
        xf86Msg(X_ERROR, "%s: All available devices are already used\n",
                pInfo->name);
    }

    return pPriv->device != NULL;
}

static int
EloPreInit(InputDriverPtr pDrv, InputInfoPtr pInfo, int flags)
{
    EloDevicePtr pPriv;

    pPriv = (EloDevicePtr) malloc(sizeof(EloDeviceRec));
    if (pPriv == NULL)
        return BadAlloc;

    memset(pPriv, 0x00, sizeof(EloDeviceRec));

    pInfo->control_proc = NULL;
    pInfo->switch_mode = NULL;
    pInfo->private = pPriv;
    pInfo->type_name = XI_TOUCHSCREEN;

    /*
     * Try to get "Device" option value from configuration file or scan "/dev/input/" directory if "Device"
     * option value is not set.
     */
    if (EloGetDevice(pInfo) == FALSE) {
        return BadValue;
    }
    else {
        SYSCALL(pInfo->fd = open(pPriv->device, O_RDWR | O_NONBLOCK));
        if (pInfo->fd < 0) {
            xf86Msg(X_ERROR, "%s: Unable to open touchscreen device %s\n",
                    pInfo->name, pPriv->device);
            return BadValue;
        }
        SYSCALL(close(pInfo->fd));
        pInfo->fd = -1;
    }

    /*
     * Parse "Device" option value and select controller connection type:
     *
     * String value like "/dev/input/event*"        - USB connection (input event interface)
     * String value like "/dev/ttyS*"           - Serial (RS232) connection
     *
     */
    if (strstr(pPriv->device, "/dev/input/event") != NULL) {
        /* USB */
        pInfo->device_control = EloUsbControl;
        pInfo->read_input = EloUsbReadInput;

        pPriv->controller = CONTROLLER_TYPE_USB;

        /* Default resolution value */

        pPriv->minX = 0;
        pPriv->minY = 0;

        pPriv->maxX = 4095;
        pPriv->maxY = 4095;

        pPriv->reverseX = FALSE;
        pPriv->reverseY = TRUE;
    }
    else if (strstr(pPriv->device, "/dev/ttyS") != NULL) {
        /* RS232 */
        pInfo->device_control = EloSerialControl;
        pInfo->read_input = EloSerialReadInput;

        pPriv->controller = CONTROLLER_TYPE_RS232;

        /* Default resolution value */

        pPriv->minX = 0;
        pPriv->minY = 0;

        pPriv->maxX = 3000;
        pPriv->maxY = 3000;

        pPriv->reverseX = FALSE;
        pPriv->reverseY = FALSE;
    }
    else {
        xf86Msg(X_ERROR, "%s: Bad \"Device\" value in config file\n",
                pInfo->name);
        return BadValue;
    }

    pPriv->minX = xf86SetIntOption(pInfo->options, "MinX", pPriv->minX);
    xf86Msg(X_INFO, "%s: minimum x position: %d\n", pInfo->name, pPriv->minX);

    pPriv->minY = xf86SetIntOption(pInfo->options, "MinY", pPriv->minY);
    xf86Msg(X_INFO, "%s: minimum y position: %d\n", pInfo->name, pPriv->minY);

    pPriv->maxX = xf86SetIntOption(pInfo->options, "MaxX", pPriv->maxX);
    xf86Msg(X_INFO, "%s: maximum x position: %d\n", pInfo->name, pPriv->maxX);

    pPriv->maxY = xf86SetIntOption(pInfo->options, "MaxY", pPriv->maxY);
    xf86Msg(X_INFO, "%s: maximum y position: %d\n", pInfo->name, pPriv->maxY);

    pPriv->width = pPriv->maxX - pPriv->minX;
    pPriv->height = pPriv->maxY - pPriv->minY;

    if (pPriv->width <= 0 || pPriv->height <= 0) {
        xf86Msg(X_ERROR,
                "%s: Can not configure touchscreen with height = %d and width = %d !\n",
                pInfo->name, pPriv->height, pPriv->width);
        return BadValue;
    }

    pPriv->screenNumber = xf86SetIntOption(pInfo->options, "ScreenNumber", 0);
    xf86Msg(X_INFO, "%s: ScreenNumber value is %d\n", pInfo->name,
            pPriv->screenNumber);

    pPriv->debug = xf86SetBoolOption(pInfo->options, "Debug", FALSE);
    if (pPriv->debug == TRUE)
        xf86Msg(X_INFO, "%s: Debug trace enabled!\n", pInfo->name);

    pPriv->threshold = xf86SetIntOption(pInfo->options, "ButtonThreshold", 1);
    if (pPriv->threshold >= 0 && pPriv->threshold < ELO_BUTTON_THRESHOLD_MAX)
        xf86Msg(X_INFO, "%s: ButtonThreshold value is %d\n", pInfo->name,
                pPriv->threshold);
    else {
        xf86Msg(X_ERROR,
                "%s: Can not configure touchscreen with ButtonThreshold value %d!\n",
                pInfo->name, pPriv->threshold);
        return BadValue;
    }

    pPriv->reverseX =
        xf86SetBoolOption(pInfo->options, "ReverseX", pPriv->reverseX);
    if (pPriv->reverseX == TRUE)
        xf86Msg(X_INFO, "%s: X-axis inversion enabled\n", pInfo->name);

    pPriv->reverseY =
        xf86SetBoolOption(pInfo->options, "ReverseY", pPriv->reverseY);
    if (pPriv->reverseY == TRUE)
        xf86Msg(X_INFO, "%s: Y-axis inversion enabled\n", pInfo->name);

    return Success;
}

static void
EloUninit(InputDriverPtr pDrv, InputInfoPtr pInfo, int flags)
{
    EloDevicePtr pPriv;

    if (pInfo == NULL)
        return;

    pPriv = (EloDevicePtr) pInfo->private;
    if (pPriv) {
        if (pPriv->device)
            free(pPriv->device);

        free(pPriv);
        pInfo->private = NULL;
    }

    xf86DeleteInput(pInfo, 0);
}

/**
 * Driver initialization information. 
 */
_X_EXPORT InputDriverRec ELO = {
    ELO_VERSION_CURRENT,    /* Driver version */
    ELO_DRIVER_NAME,        /* Driver name */
    NULL,                   /* Identify proc */
    EloPreInit,             /* Pre-init proc */
    EloUninit,              /* Un-init proc */
    NULL,                   /* Used by server. Don`t touch it. */
    0                       /* Used by server. Don`t touch it. */
};

static pointer
EloPlug(pointer module, pointer options, int *ErrMajor, int *ErrMinor)
{
    xf86AddInputDriver(&ELO, module, 0);
    return module;
}

static void
EloUnplug(pointer ptr)
{
    xf86Msg(X_INFO, "unplug touchscreen\n");
}

/*
 * Module description info.
 */
static XF86ModuleVersionInfo EloVersionRec = {
    ELO_DRIVER_NAME,         /* Name of the module */
    MODULEVENDORSTRING,      /* Vendor specific string */
    MODINFOSTRING1,          /* Constant MODINFOSTRING1 to find infoarea with a binary editor/sign tool */
    MODINFOSTRING2,          /* Constant MODINFOSTRING2 to find infoarea with a binary editor/sign tool */
    XORG_VERSION_CURRENT,    /* Contains XF86_VERSION_CURRENT */
    ELO_VERSION_MAJOR,       /* Module-specific major version */
    ELO_VERSION_MINOR,       /* Module-specific minor version */
    ELO_VERSION_PATCH,       /* Module-specific patch level */
    ABI_CLASS_XINPUT,        /* ABI class that module uses */
    ABI_XINPUT_VERSION,      /* ABI version */
    MOD_CLASS_XINPUT,        /* Module class */
    {0, 0, 0, 0}             /* Contains a digital signature of the version info structure (not currently used) */
};

/*
 * This is the entry point in the module. The name
 * is setup after the pattern <module_name>ModuleData.
 * Do not change it.
 */
_X_EXPORT XF86ModuleData eloModuleData = {
    &EloVersionRec,
    EloPlug,
    EloUnplug
};
