#include "elo_rs232.h"

static void
EloSerialPtrControl(DeviceIntPtr pDev, PtrCtrl * ctrl)
{
    /* Nothing */
}

Bool
EloSerialControl(DeviceIntPtr pDev, int mode)
{
    InputInfoPtr pInfo;
    EloDevicePtr pPriv;
    Atom buttonLabel;
    Atom axisLabels[2] = { 0, 0 };
    unsigned char map[] = { 0, 1 };

    if (pDev == NULL)
        return !Success;

    pInfo = (InputInfoPtr) pDev->public.devicePrivate;
    pPriv = (EloDevicePtr) pInfo->private;

    switch (mode) {
    case DEVICE_INIT:

        xf86Msg(X_INFO, "%s: touchscreen init\n", pInfo->name);

        if (pPriv->screenNumber >= screenInfo.numScreens ||
            pPriv->screenNumber < 0)
            pPriv->screenNumber = 0;

        pPriv->screenWidth = screenInfo.screens[pPriv->screenNumber]->width;
        pPriv->screenHeight = screenInfo.screens[pPriv->screenNumber]->height;

         /**
         *
         * ---=== Init one button ===---
         *
         */

        if (InitButtonClassDeviceStruct(pDev, 1, &buttonLabel, map) == FALSE) {
            xf86Msg(X_ERROR, "%s: InitButtonClassDeviceStruct failed\n",
                    pInfo->name);
            return !Success;
        }

        if (pDev->focus == FALSE) {
            if (InitFocusClassDeviceStruct(pDev) == FALSE) {
                xf86Msg(X_ERROR, "%s: InitFocusClassDeviceStruct failed\n",
                        pInfo->name);
                return !Success;
            }
        }

        if (InitPtrFeedbackClassDeviceStruct(pDev, EloSerialPtrControl) ==
            FALSE) {
            xf86Msg(X_ERROR, "%s: InitPtrFeedbackClassDeviceStruct failed\n",
                    pInfo->name);
            return !Success;
        }

        /**
         *
         * ---=== Init X and Y axes ===---
         *
         * Device reports motions on 2 axes in absolute coordinates.
         * Axes min and max values are reported in raw coordinates.
         * Resolution is computed roughly by the difference between
         * max and min values scaled from the approximate size
         * of the screen to fit one meter.
         */

        if (InitValuatorClassDeviceStruct(pDev,
                                          2,
                                          axisLabels,
                                          GetMotionHistorySize(),
                                          Absolute) == FALSE) {
            xf86Msg(X_ERROR, "%s: InitValuatorClassDeviceStruct failed\n",
                    pInfo->name);
            return !Success;
        }

        /* X */
        InitValuatorAxisStruct(pDev,
                               0,
                               axisLabels[0],
                               pPriv->minX,
                               pPriv->maxX,
                               9500,
                               0,       /* min resolution */
                               9500,    /* max resolution */
                               Absolute);

        /* Y */
        InitValuatorAxisStruct(pDev,
                               1,
                               axisLabels[1],
                               pPriv->minY,
                               pPriv->maxY,
                               10500,
                               0,      /* min resolution */
                               10500,   /* max resolution */
                               Absolute);

        if (pDev->focus == FALSE) {
            if (InitFocusClassDeviceStruct(pDev) == FALSE) {
                xf86Msg(X_ERROR, "%s: InitFocusClassDeviceStruct failed\n",
                        pInfo->name);
                return !Success;
            }
        }

        /* Allocate motion events buffer */
        xf86MotionHistoryAllocate(pInfo);

        break;

    case DEVICE_ON:

        xf86Msg(X_INFO, "%s: touchscreen on\n", pInfo->name);

        if (pDev->public.on == TRUE)
            break;

        if (pInfo->fd < 0) {
            /* Try to open device */
            SYSCALL(pInfo->fd = open(pPriv->device, O_RDWR | O_NONBLOCK));
            if (pInfo->fd < 0) {
                xf86Msg(X_ERROR, "%s: Unable to open touchscreen device %s\n",
                        pInfo->name, pPriv->device);
                return !Success;
            }
        }

        xf86FlushInput(pInfo->fd);
        xf86AddEnabledDevice(pInfo);
        pDev->public.on = TRUE;

        break;

        /*
         * Deactive the device. After this, the device will not emit events
         * until a subsequent DEVICE_ON. Device can be closed.
         */
    case DEVICE_OFF:

        xf86Msg(X_INFO, "%s: touchscreen close\n", pInfo->name);

        pDev->public.on = FALSE;

        if (pInfo->fd >= 0)
            xf86RemoveEnabledDevice(pInfo);

        SYSCALL(close(pInfo->fd));
        pInfo->fd = -1;
        break;

        /* 
         * Final close before server exit. This is used during server shutdown.
         * Close device and free all resources.
         */
    case DEVICE_CLOSE:
        xf86Msg(X_INFO, "%s: touchscreen close\n", pInfo->name);

        pDev->public.on = FALSE;

        if (pInfo->fd >= 0)
            xf86RemoveEnabledDevice(pInfo);

        SYSCALL(close(pInfo->fd));
        pInfo->fd = -1;

        break;

#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) * 100 + GET_ABI_MINOR(ABI_XINPUT_VERSION) >= 1901
    case DEVICE_ABORT:
        return Success;
        break;
#endif

    default:
        xf86Msg(X_ERROR, "%s: Unsupported mode = %d\n", pInfo->name, mode);
        return BadValue;
        break;

    }

    return Success;
}

Bool
EloSerialGetPacket(InputInfoPtr pInfo, EloSerialPacketBuffer * pBuffer)
{
    int ret;

    if (pInfo == NULL || pBuffer == NULL)
        return !Success;

    SYSCALL(ret = read(pInfo->fd,
                       (char *) (pBuffer->data + pBuffer->used),
                       ELO_PACKET_SIZE - pBuffer->used));

    if (ret <= 0) {
        /* 
         * ENXIO  - No such device.
         * EAGAIN - Resource temporarily unavailable, let`s try again later.
         */
        if (errno == ENXIO) {
            xf86Msg(X_ERROR, "%s: Device disappeared! Disabling driver...\n",
                    pInfo->name);
            xf86RemoveEnabledDevice(pInfo);
            close(pInfo->fd);
            pInfo->fd = -1;
        }
        else if (errno != EAGAIN)
            xf86Msg(X_ERROR, "%s: Read error: %s\n", pInfo->name,
                    strerror(errno));

        return !Success;
    }

    while (ret > 0) {
        if ((pBuffer->used == 0) && (pBuffer->data[0] != ELO_SYNC_BYTE)) {
            /* Sync with the start of a packet. */
            xf86Msg(X_ERROR,
                    "%s: Dropping one byte in an attemp to synchronize!\n",
                    pInfo->name);
            memcpy((char *) &pBuffer->data[0], (char *) &pBuffer->data[1],
                   ret - 1);
        }
        else {
            /* Compute data checksum */
            if (pBuffer->used < ELO_PACKET_SIZE - 1) {
                pBuffer->checksum =
                    pBuffer->checksum + pBuffer->data[pBuffer->used];
                pBuffer->checksum = pBuffer->checksum % 256;
            }

            pBuffer->used++;
        }

        ret--;
    }

    if (pBuffer->used == ELO_PACKET_SIZE) {
        Bool valid;

        /* Validate checksum */
        valid = (pBuffer->checksum == pBuffer->data[ELO_PACKET_SIZE - 1]);

        /* Reset packet buffer */
        memset(pBuffer->data, 0x00, ELO_PACKET_SIZE);
        pBuffer->used = 0;
        pBuffer->checksum = ELO_INIT_CHECKSUM;

        if (valid == TRUE)
            return Success;
        else
            xf86Msg(X_ERROR, "%s: checksum error!\n", pInfo->name);
    }

    return !Success;
}

void
EloSerialReadInput(InputInfoPtr pInfo)
{
    EloSerialPacketBuffer buffer;
    EloDevicePtr pPriv;
    int x;
    int y;

    pPriv = (EloDevicePtr) pInfo->private;
    if (pPriv == NULL)
        return;

    /* reset packet buffer */
    memset(&buffer.data, 0x00, ELO_PACKET_SIZE);
    buffer.used = 0;
    buffer.checksum = ELO_INIT_CHECKSUM;

    while (xf86WaitForInput(pInfo->fd, 0) > 0) {
        if (EloSerialGetPacket(pInfo, &buffer) != Success)
            continue;

        /*
         * Packet format:
         *
         * Byte 0: ELO_SYNC_BYTE
         * Byte 1: ELO_TOUCH
         *
         * Byte 2: packet type
         *  bit 2: pen up   (release)
         *  bit 1: position (stream)
         *  bit 0: pen down (press)
         *
         * Byte 3: X coordinate (lower bits)
         * Byte 4: X coordinate (upper bits)
         * Byte 5: Y coordinate (lower bits)
         * Byte 6: Y coordinate (upper bits)
         * Byte 7: Z coordinate (lower bits)
         * Byte 8: Z coordinate (upper bits)
         * Byte 9: checksum
         *
         */

        if (buffer.data[1] == ELO_TOUCH) {
            pPriv->x = WORD_ASSEMBLY(buffer.data[3], buffer.data[4]);
            pPriv->y = WORD_ASSEMBLY(buffer.data[5], buffer.data[6]);
            pPriv->z = WORD_ASSEMBLY(buffer.data[7], buffer.data[8]);

            /*
             * ( buffer.data[2] & 0x07 ) value is:
             *
             * 1. ELO_STATE_PRESS
             * 2. ELO_STATE_RELEASE
             * 3. ELO_STATE_STREAM
             *
             */

            pPriv->state = buffer.data[2] & 0x07;

            if (pPriv->reverseX == TRUE)
                x = pPriv->width - pPriv->x;
            else
                x = pPriv->x;

            if (pPriv->reverseY == TRUE)
                y = pPriv->height - pPriv->y;
            else
                y = pPriv->y;

            xf86PostMotionEvent(pInfo->dev, TRUE, 0, 2, x, y);

            if ((pPriv->state == ELO_STATE_PRESS) &&
                (pPriv->z > pPriv->threshold)) {
                xf86PostButtonEvent(pInfo->dev, TRUE, 1, TRUE, 0, 2, x, y);
            }
            else if ((pPriv->state == ELO_STATE_RELEASE) ||
                     (pPriv->z <= pPriv->threshold)) {
                xf86PostButtonEvent(pInfo->dev, TRUE, 1, FALSE, 0, 2, x, y);
            }

            if (pPriv->debug == TRUE) {
                xf86Msg(X_INFO, "%s, x = %d, y = %d, z = %d, state = %s\n",
                        pInfo->name, pPriv->x, pPriv->y, pPriv->z,
                        pPriv->state == ELO_STATE_PRESS ? "down" : "up");
            }
        }
    }
}
