/********************************************
*
* Copyright (c) 2022-2024 AO MCST
* Distributed under the terms of MIT License.
*
*********************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <xorg-server.h>
#include "xf86.h"
#include "xf86_OSproc.h"

#include "micmap.h"
#include "colormapst.h"
#include "xf86cmap.h"
#include "shadow.h"
#include "dgaproc.h"

/* for visuals */
#include "fb.h"

#include "fbdevhw.h"

#include "xf86xv.h"

#include "mcstmga.h"

#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

extern void (*SetupVideoCard)(ScrnInfoPtr pScrn, int as);

/* -------------------------------------------------------------------- */
/* prototypes                                                           */

static const OptionInfoRec *McstMgaAvailableOptions(int chipid, int busid);
static void McstMgaIdentify(int flags);
static Bool McstMgaProbe(DriverPtr drv, int flags);
static Bool McstMgaPreInit(ScrnInfoPtr pScrn, int flags);
#if ABI_VIDEODRV_VERSION < SET_ABI_VERSION(13, 0)
static Bool McstMgaScreenInit(int Index, ScreenPtr pScreen, int argc,
                              char **argv);
static Bool McstMgaCloseScreen(int scrnIndex, ScreenPtr pScreen);
#else
static Bool McstMgaScreenInit(ScreenPtr pScreen, int argc, char **argv);
static Bool McstMgaCloseScreen(ScreenPtr pScreen);
#endif
static Bool McstMgaDGAInit(ScrnInfoPtr pScrn, ScreenPtr pScreen);
static void *McstMgaWindowLinear(ScreenPtr pScreen, CARD32 row, CARD32 offset,
                                 int mode, CARD32 *size, void *closure);

static Bool McstMgaDriverFunc(ScrnInfoPtr pScrn, xorgDriverFuncOp op,
                              pointer ptr);

/* -------------------------------------------------------------------- */

#define MCSTMGA_VERSION 1000
#define MCSTMGA_NAME        "MCSTMGA"
#define MCSTMGA_DRIVER_NAME "mcstmga"

#define MCSTMGA_MAJOR_VERSION       PACKAGE_VERSION_MAJOR
#define MCSTMGA_MINOR_VERSION       PACKAGE_VERSION_MINOR
#define MCSTMGA_PATCHLEVEL          PACKAGE_VERSION_PATCHLEVEL

_X_EXPORT DriverRec McstMga = {
    MCSTMGA_VERSION,
    MCSTMGA_DRIVER_NAME,
    McstMgaIdentify,
    McstMgaProbe,
    McstMgaAvailableOptions,
    NULL,
    0,
    McstMgaDriverFunc
};

/* Supported "chipsets" */
static SymTabRec McstMgaChipsets[] = {
    {0, "mcst mga"},
    {-1, NULL}
};

/* Supported options */
typedef enum {
    OPTION_HW_CURSOR,
    OPTION_NOACCEL,
    OPTION_NODMA,
    OPTION_FBDEV,
    OPTION_SHADOW_FB,
    OPTION_ACCELMETHOD
} McstMgaOpts;

static const OptionInfoRec McstMgaOptions[] = {
    {OPTION_HW_CURSOR, "HWcursor", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_NOACCEL, "NoAccel", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_NODMA, "NoDMA", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_FBDEV, "fbdev", OPTV_STRING, {0}, FALSE},
    {OPTION_SHADOW_FB, "ShadowFB", OPTV_BOOLEAN, {0}, FALSE},
    {OPTION_ACCELMETHOD, "AccelMethod", OPTV_STRING, {0}, FALSE},
    {-1, NULL, OPTV_NONE, {0}, FALSE}
};

/* -------------------------------------------------------------------- */

#ifdef XFree86LOADER

MODULESETUPPROTO(McstMgaSetup);

static XF86ModuleVersionInfo McstMgaVersRec = {
    "mgam",
    MODULEVENDORSTRING,
    MODINFOSTRING1,
    MODINFOSTRING2,
    XORG_VERSION_CURRENT,
    MCSTMGA_MAJOR_VERSION,
    MCSTMGA_MINOR_VERSION,
    MCSTMGA_PATCHLEVEL,
    ABI_CLASS_VIDEODRV,
    ABI_VIDEODRV_VERSION,
    NULL,
    {0, 0, 0, 0}
};

_X_EXPORT
XF86ModuleData mcstmgaModuleData = { &McstMgaVersRec, McstMgaSetup, NULL };

pointer
McstMgaSetup(pointer module, pointer opts, int *errmaj, int *errmin)
{
    static Bool setupDone = FALSE;

    if (!setupDone) {
        setupDone = TRUE;
        xf86AddDriver(&McstMga, module, 0);
        return (pointer) 1;
    };
    if (errmaj)
        *errmaj = LDR_ONCEONLY;
    return NULL;
}

#endif                          /* XFree86LOADER */

/* -------------------------------------------------------------------- */
/* two functions to allocate/free private data                          */

static Bool
McstMgaAllocRec(ScrnInfoPtr pScrn)
{
    if (pScrn->driverPrivate != NULL)
        return TRUE;

    pScrn->driverPrivate = xnfcalloc(sizeof(McstMgaRec), 1);
    return TRUE;
}

static void
McstMgaFreeRec(ScrnInfoPtr pScrn)
{
    if (pScrn->driverPrivate == NULL)
        return;
    free(pScrn->driverPrivate);
    pScrn->driverPrivate = NULL;
}

static const OptionInfoRec *
McstMgaAvailableOptions(int chipid, int busid)
{
    return McstMgaOptions;
}

static void
McstMgaIdentify(int flags)
{
    xf86PrintChipsets(MCSTMGA_NAME, "driver for framebuffer", McstMgaChipsets);
}

static Bool
McstMgaProbe(DriverPtr drv, int flags)
{
    Bool foundScreen = FALSE;
    GDevPtr *devSections;
    int numDevSections;
    ScrnInfoPtr pScrn;
    char *dev;
    int i;

    /* For now, just bail out for PROBE_DETECT. */
    if (flags & PROBE_DETECT)
        return FALSE;

    /*
     * Find the config file Device sections that match this
     * driver, and return if there are none.
     */
    if ((numDevSections =
         xf86MatchDevice(MCSTMGA_DRIVER_NAME, &devSections)) <= 0)
        return FALSE;

    if (!xf86LoadDrvSubModule(drv, "ramdac"))
        return FALSE;

    if (!xf86LoadDrvSubModule(drv, "fbdevhw"))
        return FALSE;

    for (i = 0; i < numDevSections; i++) {
        dev = (char *) xf86FindOptionValue(devSections[i]->options, "fbdev");

        if (fbdevHWProbe(NULL, dev, NULL)) {
            int entity;
            pScrn = NULL;

            entity = xf86ClaimFbSlot(drv, 0, devSections[i], TRUE);
            pScrn = xf86ConfigFbEntity(pScrn, 0, entity,
                                       NULL, NULL, NULL, NULL);
            if (pScrn) {
                foundScreen = TRUE;

                pScrn->driverVersion = MCSTMGA_VERSION;
                pScrn->driverName = MCSTMGA_DRIVER_NAME;
                pScrn->name = MCSTMGA_NAME;
                pScrn->Probe = McstMgaProbe;
                pScrn->PreInit = McstMgaPreInit;
                pScrn->ScreenInit = McstMgaScreenInit;
                pScrn->SwitchMode = fbdevHWSwitchModeWeak();
                pScrn->AdjustFrame = fbdevHWAdjustFrameWeak();
                pScrn->EnterVT = fbdevHWEnterVTWeak();
                pScrn->LeaveVT = fbdevHWLeaveVTWeak();
                pScrn->ValidMode = fbdevHWValidModeWeak();

                xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                           "using %s\n", dev ? dev : "default device");
            }
        }
    }
    free(devSections);
    return foundScreen;
}

char *optstr = NULL;

static Bool
McstMgaPreInitAccel(ScrnInfoPtr pScrn)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    MessageType from;
    int errmaj = 0, errmin = 0;

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_DEFAULT, "McstMgaPreInitAccel start\n");
#endif
#ifdef USE_XAA
    infoPtr->useXAA = FALSE;
#endif /* USE_XAA */
    if (xf86ReturnOptValBool(infoPtr->Options, OPTION_NOACCEL, FALSE)) {
        infoPtr->NoAccel = TRUE;
        infoPtr->useEXA = FALSE;
        return (FALSE);
    };
    if (xf86ReturnOptValBool(infoPtr->Options, OPTION_NODMA, FALSE))
        infoPtr->NoDMA = TRUE;
    else
        infoPtr->NoDMA = FALSE;

    from = X_DEFAULT;
    infoPtr->NoAccel = FALSE;
    infoPtr->useEXA = TRUE;
    optstr = (char *) xf86GetOptValString(infoPtr->Options, OPTION_ACCELMETHOD);
    if (optstr != NULL) {
        xf86DrvMsg(pScrn->scrnIndex, X_DEFAULT, "Opt string %s\n", optstr);
        if (xf86NameCmp(optstr, "EXA") == 0) {
            infoPtr->useEXA = TRUE;
#ifdef USE_XAA
            infoPtr->useXAA = FALSE;
        }
        else if (xf86NameCmp(optstr, "XAA") == 0) {
            infoPtr->useEXA = FALSE;
            infoPtr->useXAA = TRUE;
#endif /* USE_XAA */
        }
        from = X_CONFIG;
    }
    if (infoPtr->useEXA)
        ErrorF("Using EXA acceleration architecture\n");
#ifdef USE_XAA
    else
        ErrorF("Using XAA acceleration architecture\n");
#endif /* USE_XAA */

#ifdef XFree86LOADER
    infoPtr->exaReq.majorversion = EXA_VERSION_MAJOR;
    infoPtr->exaReq.minorversion = EXA_VERSION_MINOR;
    if (!LoadSubModule(pScrn->module, "exa", NULL, NULL, NULL,
                       &infoPtr->exaReq, &errmaj, &errmin)) {
        LoaderErrorMsg(NULL, "exa", errmaj, errmin);
        return FALSE;
    };

#ifdef USE_XAA
    if (!xf86LoadSubModule(pScrn, "xaa"))
        return FALSE;
#endif /* USE_XAA */
#endif
    return TRUE;
}

static Bool
McstMgaPreInit(ScrnInfoPtr pScrn, int flags)
{
    MessageType from = X_DEFAULT;
    ClockRangePtr clockRanges;  /* RAMDAC clocks range */
    int depth24Flags = 0;
    McstMgaPtr infoPtr;
    int defaultDepth;
    int modesFound;             /* number of matching modes found */
    int fbbpp;

    if (flags & PROBE_DETECT)
        return FALSE;

    /* Check the number of entities, and fail if it isn't one. */
    if (pScrn->numEntities != 1)
        return FALSE;

    pScrn->monitor = pScrn->confScreen->monitor;

    McstMgaAllocRec(pScrn);
    infoPtr = McstMgaGetRec(pScrn);
    infoPtr->NoAccel = TRUE;
    infoPtr->useEXA = FALSE;
#ifdef USE_XAA
    infoPtr->useXAA = FALSE;
#endif /* USE_XAA */
    infoPtr->pEnt = xf86GetEntityInfo(pScrn->entityList[0]);

#ifndef XSERVER_LIBPCIACCESS
    pScrn->racMemFlags = RAC_FB | RAC_COLORMAP | RAC_CURSOR | RAC_VIEWPORT;
    pScrn->racIoFlags = RAC_FB | RAC_COLORMAP | RAC_CURSOR | RAC_VIEWPORT;
#endif

    /* open device */
    if (!fbdevHWInit
        (pScrn, NULL,
         (char *) xf86FindOptionValue(infoPtr->pEnt->device->options,
                                      (char *) "fbdev")))
        return FALSE;

    /* Check for valid Linux framebuffer device driver for MGA video card */
    if (strcmp(fbdevHWGetName(pScrn), "mcst_mga") != 0) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "invalid Linux framebuffer device driver (%s ?)\n",
                   fbdevHWGetName(pScrn));
        return FALSE;
    }

    defaultDepth = fbdevHWGetDepth(pScrn, &fbbpp);

    depth24Flags = Support24bppFb | Support32bppFb
        | SupportConvert24to32 | PreferConvert24to32;

    if (!xf86SetDepthBpp(pScrn, 0, 0, 0, depth24Flags))
        return FALSE;

    xf86PrintDepthBpp(pScrn);

    /* color weight */
    if (pScrn->depth > 8) {
        rgb zeros = { 0, 0, 0 };
        if (!xf86SetWeight(pScrn, zeros, zeros))
            return FALSE;
    }

    /* visual init */
    if (!xf86SetDefaultVisual(pScrn, -1))
        return FALSE;

    /* We don't currently support DirectColor at > 8bpp */
    if (pScrn->depth > 8 && pScrn->defaultVisual != TrueColor) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Given default visual"
                   " (%s) is not supported at depth %d\n",
                   xf86GetVisualName(pScrn->defaultVisual), pScrn->depth);
        return FALSE;
    }

    {
        Gamma zeros = { 0.0, 0.0, 0.0 };

        if (!xf86SetGamma(pScrn, zeros))
            return FALSE;
    }

    pScrn->progClock = TRUE;
    pScrn->rgbBits = 8;
    pScrn->chipset = "mcst mga";
    pScrn->videoRam = fbdevHWGetVidmem(pScrn) / 1024;   /* in Kbytes */

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Hardware: %s (vidmem: %dKiB)\n",
               fbdevHWGetName(pScrn), pScrn->videoRam);
#endif

    /* handle options */
    xf86CollectOptions(pScrn, NULL);
    if (!(infoPtr->Options = malloc(sizeof(McstMgaOptions))))
        return FALSE;
    memcpy(infoPtr->Options, McstMgaOptions, sizeof(McstMgaOptions));
    xf86ProcessOptions(pScrn->scrnIndex, infoPtr->pEnt->device->options,
                       infoPtr->Options);

    /* use shadow framebuffer by default */
    if(xf86ReturnOptValBool(infoPtr->Options, OPTION_SHADOW_FB, FALSE)) {
#if XORG_VERSION_CURRENT > XORG_VERSION_1_12_2
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Option ShadowFB is available \
with xorg-server version <= 1.12.2\n");
        infoPtr->shadowFB = FALSE;
#else
        infoPtr->shadowFB = TRUE;
#endif
    }
    else
        infoPtr->shadowFB = FALSE;

    /* cursor type */
    from = X_DEFAULT;
    infoPtr->HWCursor = TRUE;
    if (xf86GetOptValBool
        (infoPtr->Options, OPTION_HW_CURSOR, &infoPtr->HWCursor))
        from = X_CONFIG;
    xf86DrvMsg(pScrn->scrnIndex, from, "Using %s cursor\n",
               infoPtr->HWCursor ? "HW" : "SW");

    /* Acceleration pre-init */
    McstMgaPreInitAccel(pScrn);

    /* xf86ValidateModes take care of clockRanges */
    clockRanges = xnfcalloc(sizeof(ClockRange), 1);
    clockRanges->next = NULL;
    clockRanges->minClock = 0;
    clockRanges->maxClock = 185000;     /* 185Mhz */
    clockRanges->clockIndex = -1;       /* programmable */
    clockRanges->interlaceAllowed = FALSE;
    clockRanges->doubleScanAllowed = FALSE;

    /* select video modes */
    modesFound = xf86ValidateModes(pScrn,
                                   pScrn->monitor->Modes,
                                   pScrn->display->modes,
                                   clockRanges,
                                   NULL, 0, 65535, pScrn->bitsPerPixel,
                                   0, 65535,
                                   pScrn->display->virtualX,
                                   pScrn->display->virtualY,
                                   (pScrn->videoRam * 1024) -
                                   MCSTMGA_CURSOR_SIZE, LOOKUP_DEFAULT);

    if (modesFound == -1) {
        McstMgaFreeRec(pScrn);
        return FALSE;
    }

    if (modesFound < 1 && infoPtr->fbmem) {
        fbdevHWUseBuildinMode(pScrn);
        pScrn->displayWidth =
            fbdevHWGetLineLength(pScrn) / (pScrn->bitsPerPixel >> 3);
        modesFound = 1;
    }

    /* Prune the modes marked as invalid */
    xf86PruneDriverModes(pScrn);

    if (modesFound == 0 || pScrn->modes == NULL) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "No valid modes found\n");
        McstMgaFreeRec(pScrn);
        return FALSE;
    }

    /*
     * Set the CRTC parameters for all of the modes based on the type
     * of mode, and the chipset's interlace requirements.
     *
     * Calling this is required if the mode->Crtc* values are used by the
     * driver and if the driver doesn't provide code to set them.  They
     * are not pre-initialised at all.
     */
    xf86SetCrtcForModes(pScrn, 0);

    /* Set the current mode to the first in the list */
    pScrn->currentMode = pScrn->modes;

    /* Print the list of modes being used */
    xf86PrintModes(pScrn);

    /* Set display resolution */
    xf86SetDpi(pScrn, 0, 0);

    /* Load bpp-specific modules */
    if (fbdevHWGetType(pScrn) == FBDEVHW_PACKED_PIXELS) {
        switch (pScrn->bitsPerPixel) {
        case 8:
        case 16:
        case 24:
        case 32:
            if (xf86LoadSubModule(pScrn, "fb") == NULL) {
                McstMgaFreeRec(pScrn);
                return FALSE;
            };
            break;
        default:
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Unsupported bpp: %d", pScrn->bitsPerPixel);
            McstMgaFreeRec(pScrn);
            return FALSE;
        }
    }
    else {
        /* This should never happen ...
         * we should check for this much much earlier ... */
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "invalid fbdev type (%d).\n", fbdevHWGetType(pScrn));
        McstMgaFreeRec(pScrn);
        return FALSE;
    }

    /* Load shadow if needed */
    if (infoPtr->shadowFB) {
        xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
                   "Using \"Shadow Framebuffer\"\n");
        if (!xf86LoadSubModule(pScrn, "shadow")) {
            McstMgaFreeRec(pScrn);
            return FALSE;
        };
    }

    return TRUE;
}

void
mmioRead(int fbNum)
{
    char str[32];
    int *fb;
    int fd;
    int i;
    int k;

    sprintf(str, "/dev/fb%d", fbNum);
    fd = open(str, O_RDWR);
    if (fd == -1) {
        ErrorF("Can not open frame buffer: %s\n", str);
        return;
    };
    fb = mmap(0, 0x0001000, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0x800000);
    if (fb == (int *) -1) {
        ErrorF("Can not mmap() frame buffer: %s\n", str);
        close(fd);
        return;
    };
    k = 0;
    for (i = 0; i < 256; i = i + 4) {
        ErrorF("%4.4X: %8.8X   ", i, fb[i >> 2]);
        if (k++ > 3) {
            ErrorF("\n");
            k = 0;
        }
    };
    ErrorF("\n");
    close(fd);
    return;
}

extern int xf86NumScreens;

void
McstMgaSetupRegs(ScrnInfoPtr pSc, int as)
/* as = 1 - came back from terminal mode */
/* as = 0 - beginning set up */
{
    McstMgaPtr infoPtr;
    ScrnInfoPtr pScrn;
    CARD32 r;
    CARD32 i;
    CARD32 n;

    if (as == 0)
        n = 1;
    else
        n = xf86NumScreens;
    for (i = 0; i < n; i++) {
        if (pSc == (ScrnInfoPtr) NULL)
            pScrn = xf86Screens[i];
        else
            pScrn = pSc;
        infoPtr = McstMgaGetRec(pScrn);
        r = McstMgaMmio32Read(infoPtr, REG_CTRL);
        r = (r & CTRL_FLIP_MASK & CTRL_CD_MASK);
        switch (pScrn->bitsPerPixel) {
        case 8:
            break;
        case 16:
            r = r | CTRL_FLIP_BYTES | CTRL_CD_16BPP;
            break;
        case 24:
            r = r | CTRL_FLIP_BYTES | CTRL_FLIP_HWORD | CTRL_CD_24BPP;
            break;
        case 32:
            r = r | CTRL_FLIP_BYTES | CTRL_FLIP_HWORD | CTRL_CD_32BPP;
            break;
        };
        McstMgaMmio32Write(infoPtr, REG_CTRL, r);
    }
}

/**
 * Loads palette directly to video card registers
 */
static void
McstMgaLoadPalette(ScrnInfoPtr pScrn, int numColors, int *indices,
                   LOCO * colors, VisualPtr pVisual)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int i;

    for (i = 0; i < numColors; i++) {
        CARD32 val = (colors[indices[i]].red & 0xFF) << 16
            | (colors[indices[i]].green & 0xFF) << 8
            | (colors[indices[i]].blue & 0xFF);
        McstMgaMmio32Write(infoPtr, REG_PCLT + indices[i] * 4, val);
    };
}

#if ABI_VIDEODRV_VERSION < SET_ABI_VERSION(13, 0)
static Bool
McstMgaScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
#else
static Bool
McstMgaScreenInit(ScreenPtr pScreen, int argc, char **argv)
#endif
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    VisualPtr visual;
    int ret;
    int width;
    int height;
    int type;

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
               "\tbitsPerPixel=%d, depth=%d, defaultVisual=%s\n"
               "\tmask: %lx,%lx,%lx, offset: %ld,%ld,%ld\n",
               pScrn->bitsPerPixel, pScrn->depth,
               xf86GetVisualName(pScrn->defaultVisual), pScrn->mask.red,
               pScrn->mask.green, pScrn->mask.blue, pScrn->offset.red,
               pScrn->offset.green, pScrn->offset.blue);
#endif

    if (NULL == (infoPtr->fbmem = fbdevHWMapVidmem(pScrn))) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Map vid mem failed\n");
        return FALSE;
    }
    infoPtr->fbmem_len = fbdevHWGetVidmem(pScrn);
    infoPtr->fboff = fbdevHWLinearOffset(pScrn);

    if (NULL == (infoPtr->fbmmio = fbdevHWMapMMIO(pScrn)))
        return FALSE;
#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "infoPtr: fbmem 0x%lx fbmmio 0x%lx"
               "pScrn: memPhysBase 0x%lx, fbOffset 0x%lx",
               (unsigned long)infoPtr->fbmem, (unsigned long)infoPtr->fbmmio,
               pScrn->memPhysBase, pScrn->fbOffset);
#endif /* MCSTMGA_DEBUG */

    {
        /* Reading firmware version */
        CARD32 stamp = McstMgaMmio32Read(infoPtr, REG_STAMP);
        infoPtr->firmwareVersion = (stamp & 0xff000000) >> 24;
        infoPtr->firmwareDate = stamp & 0x00ffffff;

        if (serverGeneration == 1) {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                       "Firmware: version %d (20%02d-%02d-%02d)\n",
                       infoPtr->firmwareVersion,
                       (infoPtr->firmwareDate & 0xff0000) >> 16,
                       (infoPtr->firmwareDate & 0xff00) >> 8,
                       (infoPtr->firmwareDate & 0xff));
        }
    }

    fbdevHWSave(pScrn);

    if (!fbdevHWModeInit(pScrn, pScrn->currentMode)) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Mode init failed\n");
        return FALSE;
    }
    fbdevHWSaveScreen(pScreen, SCREEN_SAVER_ON);

    /* mi layer */
    miClearVisualTypes();
    if (pScrn->bitsPerPixel > 8) {
        if (!miSetVisualTypes
            (pScrn->depth, TrueColorMask, pScrn->rgbBits, TrueColor)) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Set visual types failed\n");
            return FALSE;
        }
    }
    else {
        if (!miSetVisualTypes(pScrn->depth,
                              miGetDefaultVisualMask(pScrn->depth),
                              pScrn->rgbBits, pScrn->defaultVisual)) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Set visual types failed\n");
            return FALSE;
        }
    }
    if (!miSetPixmapDepths()) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Set pixmap depths failed\n");
        return FALSE;
    }

    width = pScrn->virtualX;
    height = pScrn->virtualY;

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "shadowFB\n");
#endif  /* MCSTMGA_DEBUG */
    /* shadowfb */
    if (infoPtr->shadowFB) {
        if ((infoPtr->shadowmem = shadowAlloc(width, height,
                                              pScrn->bitsPerPixel)) == NULL) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Allocation of shadow memory failed\n");
            return FALSE;
        }
        infoPtr->fbstart = infoPtr->shadowmem;
    }
    else {
        infoPtr->shadowmem = NULL;
        infoPtr->fbstart = infoPtr->fbmem + infoPtr->fboff;
    }

    if ((type = fbdevHWGetType(pScrn)) == FBDEVHW_PACKED_PIXELS) {
        switch (pScrn->bitsPerPixel) {
        case 8:
        case 16:
        case 24:
        case 32:
            ret = fbScreenInit(pScreen, infoPtr->fbstart, width, height,
                               pScrn->xDpi, pScrn->yDpi, pScrn->displayWidth,
                               pScrn->bitsPerPixel);
            break;
        default:
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Internal error: invalid bpp (%d) in McstMgaScreenInit\n",
                       pScrn->bitsPerPixel);
            ret = FALSE;
            break;
        }
    }
    else {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Internal error: fbdev type (%d) unsupported in McstMgaScreenInit\n",
                   type);
        ret = FALSE;
    }

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ret\n");
#endif /* MCSTMGA_DEBUG */
    if (!ret)
        return FALSE;

    if (pScrn->bitsPerPixel > 8) {
        /* Fixup RGB ordering */
        visual = pScreen->visuals + pScreen->numVisuals;
        while (--visual >= pScreen->visuals) {
            if ((visual->class | DynamicClass) == DirectColor) {
                visual->offsetRed = pScrn->offset.red;
                visual->offsetGreen = pScrn->offset.green;
                visual->offsetBlue = pScrn->offset.blue;
                visual->redMask = pScrn->mask.red;
                visual->greenMask = pScrn->mask.green;
                visual->blueMask = pScrn->mask.blue;
            }
        }
    }

    /* must be after RGB ordering fixed */
    if (!fbPictureInit(pScreen, NULL, 0))
        xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
                   "RENDER extension initialisation failed.\n");

    if (infoPtr->shadowFB
        && !shadowInit(pScreen, shadowUpdatePackedWeak(), McstMgaWindowLinear))
    {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Shadow framebuffer initialization failed.\n");
        return FALSE;
    }

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "DGAInit\n");
#endif /* MCSTMGA_DEBUG */
    McstMgaDGAInit(pScrn, pScreen);

    xf86SetBlackWhitePixels(pScreen);
    xf86SetBackingStore(pScreen);

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Accel init\n");
#endif /* MCSTMGA_DEBUG */
    /* Initialize hardware 2D acceleration */
    if (!infoPtr->shadowFB && !infoPtr->NoAccel) {
        if (McstMgaAccelInit(pScreen)) {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration enabled\n");
        }
        else {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Acceleration initialization failed\n");
            xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration disabled\n");
        }
    }
    else {
        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Acceleration disabled\n");
    }

    /* software cursor */
    miDCInitialize(pScreen, xf86GetPointerScreenFuncs());

    /*
     * Initialize HW cursor layer. Must be after software cursor initialization
     */
    if (infoPtr->HWCursor) {
        if (!McstMgaHWCursorInit(pScreen))
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Hardware cursor initialization failed.\n");
    }

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "ColorMap\n");
#endif /* MCSTMGA_DEBUG */

    /* colormap */
    if ((type = fbdevHWGetType(pScrn)) == FBDEVHW_PACKED_PIXELS) {
        if (!miCreateDefColormap(pScreen))
            return FALSE;
    }
    else {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Internal error: invalid fbdev type (%d) in McstMgaScreenInit\n",
                   type);
        return FALSE;
    }

    if (!xf86HandleColormaps(pScreen, 256, 8, McstMgaLoadPalette,
                             NULL, CMAP_PALETTED_TRUECOLOR)) {
        return FALSE;
    };

    xf86DPMSInit(pScreen, fbdevHWDPMSSetWeak(), 0);

    pScreen->SaveScreen = fbdevHWSaveScreenWeak();

#ifdef MCSTMGA_DEBUG
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "Close screen\n");
#endif /* MCSTMGA_DEBUG */
    /* Wrap the current CloseScreen function */
    infoPtr->CloseScreen = pScreen->CloseScreen;
    pScreen->CloseScreen = McstMgaCloseScreen;

    {
        XF86VideoAdaptorPtr *ptr;

        int n = xf86XVListGenericAdaptors(pScrn, &ptr);
        if (n) {
            xf86XVScreenInit(pScreen, ptr, n);
        }
    }
#ifdef __E2K__
    SetupVideoCard = McstMgaSetupRegs;
    McstMgaSetupRegs(pScrn, 0);
#endif

    return TRUE;
}

#if ABI_VIDEODRV_VERSION < SET_ABI_VERSION(13, 0)
static Bool
McstMgaCloseScreen(int scrnIndex, ScreenPtr pScreen)
#else
static Bool
McstMgaCloseScreen(ScreenPtr pScreen)
#endif
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    if (infoPtr->CursorInfoRec) {
        xf86CursorInfoPtr cursorPtr = infoPtr->CursorInfoRec;
        if (cursorPtr->HideCursor)
            cursorPtr->HideCursor(pScrn);
        xf86DestroyCursorInfoRec(infoPtr->CursorInfoRec);
    }

    if (!infoPtr->NoAccel) {
        McstMgaAccelCleanup(pScreen);
    }

    fbdevHWRestore(pScrn);
    fbdevHWUnmapMMIO(pScrn);
    fbdevHWUnmapVidmem(pScrn);

    if (infoPtr->shadowmem) {
        free(infoPtr->shadowmem);
        infoPtr->shadowmem = NULL;
    }

    if (infoPtr->pDGAMode) {
        free(infoPtr->pDGAMode);
        infoPtr->pDGAMode = NULL;
        infoPtr->nDGAMode = 0;
    }
    pScrn->vtSema = FALSE;

    pScreen->CloseScreen = infoPtr->CloseScreen;
#if ABI_VIDEODRV_VERSION < SET_ABI_VERSION(13, 0)
    return (*pScreen->CloseScreen) (scrnIndex, pScreen);
#else
    return (*pScreen->CloseScreen) (pScreen);
#endif
}

/***********************************************************************
 * ShadowFB stuff
 ***********************************************************************/
static void *
McstMgaWindowLinear(ScreenPtr pScreen, CARD32 row, CARD32 offset, int mode,
                    CARD32 *size, void *closure)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    if (!pScrn->vtSema)
        return NULL;

    if (infoPtr->lineLength)
        *size = infoPtr->lineLength;
    else
        *size = infoPtr->lineLength = fbdevHWGetLineLength(pScrn);

    return ((CARD8 *) infoPtr->fbmem + infoPtr->fboff +
            row * infoPtr->lineLength + offset);
}

/***********************************************************************
 * DGA stuff
 ***********************************************************************/
static Bool McstMgaDGAOpenFramebuffer(ScrnInfoPtr pScrn, char **DeviceName,
                                      unsigned char **ApertureBase,
                                      int *ApertureSize, int *ApertureOffset,
                                      int *flags);
static Bool McstMgaDGASetMode(ScrnInfoPtr pScrn, DGAModePtr pDGAMode);
static void McstMgaDGASetViewport(ScrnInfoPtr pScrn, int x, int y, int flags);

static Bool
McstMgaDGAOpenFramebuffer(ScrnInfoPtr pScrn, char **DeviceName,
                          unsigned char **ApertureBase, int *ApertureSize,
                          int *ApertureOffset, int *flags)
{
    *DeviceName = NULL;         /* No special device */
    *ApertureBase = (unsigned char *) (pScrn->memPhysBase);
    *ApertureSize = pScrn->videoRam * 1024;
    *ApertureOffset = pScrn->fbOffset;
    *flags = 0;
    return TRUE;
}

static Bool
McstMgaDGASetMode(ScrnInfoPtr pScrn, DGAModePtr pDGAMode)
{
    int scrnIdx = pScrn->pScreen->myNum;
    DisplayModePtr pMode;
    int frameX0;
    int frameY0;

    if (pDGAMode) {
        pMode = pDGAMode->mode;
        frameX0 = frameY0 = 0;
    }
    else {
        if (!(pMode = pScrn->currentMode))
            return TRUE;
        frameX0 = pScrn->frameX0;
        frameY0 = pScrn->frameY0;
    };

#if ABI_VIDEODRV_VERSION < SET_ABI_VERSION(13, 0)
    if (!(*pScrn->SwitchMode) (scrnIdx, pMode, 0))
        return FALSE;
    (*pScrn->AdjustFrame) (scrnIdx, frameX0, frameY0, 0);
#else
    if (!(*pScrn->SwitchMode) (pScrn, pMode))
        return FALSE;
    (*pScrn->AdjustFrame) (pScrn, frameX0, frameY0);
#endif
    return TRUE;
}

static void
McstMgaDGASetViewport(ScrnInfoPtr pScrn, int x, int y, int flags)
{
#if ABI_VIDEODRV_VERSION < SET_ABI_VERSION(13, 0)
    (*pScrn->AdjustFrame) (pScrn->pScreen->myNum, x, y, flags);
#else
    (*pScrn->AdjustFrame) (pScrn, x, y);
#endif
}

static int
McstMgaDGAGetViewport(ScrnInfoPtr pScrn)
{
    return (0);
}

static DGAFunctionRec McstMgaDGAFunctions = {
    McstMgaDGAOpenFramebuffer,
    NULL,       /* CloseFramebuffer */
    McstMgaDGASetMode,
    McstMgaDGASetViewport,
    McstMgaDGAGetViewport,
    NULL,       /* Sync */
    NULL,       /* FillRect */
    NULL,       /* BlitRect */
    NULL,       /* BlitTransRect */
};

static void
McstMgaDGAAddModes(ScrnInfoPtr pScrn)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    DisplayModePtr pMode = pScrn->modes;
    DGAModePtr pDGAMode;

    do {
        pDGAMode = realloc(infoPtr->pDGAMode,
                           (infoPtr->nDGAMode + 1) * sizeof(DGAModeRec));
        if (!pDGAMode)
            break;

        infoPtr->pDGAMode = pDGAMode;
        pDGAMode += infoPtr->nDGAMode;
        (void) memset(pDGAMode, 0, sizeof(DGAModeRec));

        ++infoPtr->nDGAMode;
        pDGAMode->mode = pMode;
        pDGAMode->flags = DGA_CONCURRENT_ACCESS | DGA_PIXMAP_AVAILABLE;
        pDGAMode->byteOrder = pScrn->imageByteOrder;
        pDGAMode->depth = pScrn->depth;
        pDGAMode->bitsPerPixel = pScrn->bitsPerPixel;
        pDGAMode->red_mask = pScrn->mask.red;
        pDGAMode->green_mask = pScrn->mask.green;
        pDGAMode->blue_mask = pScrn->mask.blue;
        pDGAMode->visualClass = pScrn->bitsPerPixel > 8 ?
            TrueColor : PseudoColor;
        pDGAMode->xViewportStep = 1;
        pDGAMode->yViewportStep = 1;
        pDGAMode->viewportWidth = pMode->HDisplay;
        pDGAMode->viewportHeight = pMode->VDisplay;

        if (infoPtr->lineLength)
            pDGAMode->bytesPerScanline = infoPtr->lineLength;
        else
            pDGAMode->bytesPerScanline = infoPtr->lineLength =
                fbdevHWGetLineLength(pScrn);

        pDGAMode->imageWidth = pMode->HDisplay;
        pDGAMode->imageHeight = pMode->VDisplay;
        pDGAMode->pixmapWidth = pDGAMode->imageWidth;
        pDGAMode->pixmapHeight = pDGAMode->imageHeight;
        pDGAMode->maxViewportX = pScrn->virtualX - pDGAMode->viewportWidth;
        pDGAMode->maxViewportY = pScrn->virtualY - pDGAMode->viewportHeight;

        pDGAMode->address = infoPtr->fbstart;

        pMode = pMode->next;
    } while (pMode != pScrn->modes);
}

static Bool
McstMgaDGAInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    if (pScrn->depth < 8)
        return FALSE;

    if (!infoPtr->nDGAMode)
        McstMgaDGAAddModes(pScrn);

    return (DGAInit(pScreen, &McstMgaDGAFunctions,
                    infoPtr->pDGAMode, infoPtr->nDGAMode));
}

static Bool
McstMgaDriverFunc(ScrnInfoPtr pScrn, xorgDriverFuncOp op, pointer ptr)
{
    xorgHWFlags *flag;

    switch (op) {
    case GET_REQUIRED_HW_INTERFACES:
        flag = (CARD32 *)ptr;
        (*flag) = 0;
        return TRUE;
    default:
        return FALSE;
    }
}

