/********************************************
*
* Copyright (c) 2022-2024 AO MCST
* Distributed under the terms of MIT License.
*
*********************************************/

#ifndef __MCSTMGA_H
#define __MCSTMGA_H

#include <xorg-server.h>
#include "compiler.h"
#include "xf86Cursor.h"

#define XORG_VERSION_1_12_2 (((1) * 10000000) + ((12) * 100000) \
                             + ((2) * 1000) + 0)

#ifdef USE_XAA
#include "xaa.h"
#endif /* USE_XAA */

#include "exa.h"

#include "mcstmga_regs.h"

/*******************************************************************************
 * Debug
 *******************************************************************************
 */

/*
#define MCSTMGA_DEBUG
#define MCSTMGA_DEBUG_ACCEL
#define MCSTMGA_DEBUG_CURSOR
*/

#ifdef MCSTMGA_DEBUG
# define TRACE_ENTER(str)       ErrorF("mcstmga: " str " %d\n",pScrn->scrnIndex)
# define TRACE_EXIT(str)        ErrorF("mcstmga: " str " done at %d\n", __LINE__)
# define TRACE(str)             ErrorF("mcstmga trace: " str "\n")
#else
# define TRACE_ENTER(str)
# define TRACE_EXIT(str)
# define TRACE(str)
#endif

#define TRACEIN
#define TRACEOUT

/**	ioctl requests		*/

#define FBIOALLOC_DMA_MEM	    0x4631
#define FBIOFREE_ALL_DMA_MEMS	0x4632
#define FBIOWHOAMI		        0x4633

/* Control Register REG_CTRL */
#if 0
/* cursor 1 not supported by hardware */
#define CTRL_HC1R_32	    0               /* Hardware Cursor1 Resolution 32x32    */
#define	CTRL_HC1R_64	    (0x1<<25)       /*                             64x64    */
#define CTRL_HC1E	        (0x1<<24)       /* Hardware Cursor1 Enabled     */
#endif

#define	CTRL_FLIP_HWORD	    (0x1<<31)       /* Flip half words              */
#define	CTRL_FLIP_BYTES	    (0x1<<30)       /* Flip bytes in half words     */
#define	CTRL_FLIP_BITS	    (0x1<<29)       /* Flip bits in byte            */
#define	CTRL_FLIP_MASK	    0x1FFFFFFF      /* Flip mask                    */

#define	CTRL_HC0R_32	    0               /* HW Cursor0 Resolution 32x32  */
#define	CTRL_HC0R_64	    (0x1<<21)       /* HW Cursor0 Resolution 64x64  */
#define CTRL_HC0E	        (0x1<<20)       /* Hardware Cursor0 Enabled     */

#define CTRL_TST	        (0x1<<17)
#define CTRL_BL_POS	        0               /* Blanking Polarization Level Positive */
#define CTRL_BL_NEG	        (0x1<<15)       /*                             Negative */
#define CTRL_CSYNC_HIGH	    0               /* Composite Synchronization Pulse Polarization Level Positive  */
#define CTRL_CSYNC_LOW	    (0x1<<14)       /*                                                    Negative  */
#define CTRL_VSYNC_HIGH	    0               /* Vertical Synchronization Pulse Polarization Level Positive   */
#define CTRL_VSYNC_LOW	    (0x1<<13)       /*                                                   Negative   */
#define CTRL_HSYNC_HIGH	    0               /* Horizontal Synchronization Pulse polarization Level Positive */
#define CTRL_HSYNC_LOW	    (0x1<<12)       /*                                                     Negative */
#define CTRL_PC_GRAY	    0               /* 8-bit Pseudo Color Grayscale */
#define CTRL_PC_PSEUDO	    (0x1<<11)       /*                    Pseudo Color */
#define CTRL_CD_8BPP	    0               /* Color Depth  8bpp    */
#define CTRL_CD_16BPP	    (0x1<<9)        /*             16bpp    */
#define CTRL_CD_24BPP	    (0x2<<9)        /*             24bpp    */
#define CTRL_CD_32BPP	    (0x3<<9)        /*             32bpp    */
#define	CTRL_CD_MASK	    0xFFFFF9FF      /* Flip mask mask       */

#define CTRL_VBL_1	        0               /* Video Memory Burst Length 1 cycle    */
#define CTRL_VBL_2	        (0x1<<7)        /*                           2 cycles   */
#define CTRL_VBL_4	        (0x2<<7)        /*                           4 cycles   */
#define CTRL_VBL_8	        (0x3<<7)        /*                           8 cycles   */
#define CTRL_VBL1024	    (0x203<<7)      /*                          16 cycles (extension)       */

#define CTRL_CBSWE	        (0x1<<6)        /* CLUT Bank Switching Enable   */
#define CTRL_VBSWE	        (0x1<<5)        /* Video Bank Switching Enable  */
#define CTRL_CBSIE	        (0x1<<4)        /* CLUT Bank Switch Interrupt Enable    */
#define CTRL_VBSIE	        (0x1<<3)        /* VideoBank Switch Interrupt Enable    */
#define CTRL_HIE	        (0x1<<2)        /* HSync Interrupt Enable       */
#define CTRL_VIE	        (0x1<<1)        /* VSync Interrupt Enable       */
#define CTRL_VEN	        (0x1<<0)        /* Video Enable         */

/* Status Register REG_STAT */
#define STAT_HC1A	(0x1<<24)       /* Hardware cursor1 available   */
#define STAT_HC0A	(0x1<<20)       /* Hardware cursor0 available   */
#define STAT_ACMP	(0x1<<17)       /* Active CLUT Memory Page      */
#define STAT_AVMP	(0x1<<16)       /* Active Video Memory Page     */
#define STAT_CBSINT	(0x1<<7)        /* CLUT Bank Switch Interrupt Pending   */
#define STAT_VBSINT	(0x1<<6)        /* Bank Switch Interrupt Pending        */
#define STAT_HINT	(0x1<<5)        /* Horizontal Interrupt Pending */
#define STAT_VINT	(0x1<<4)        /* Vertical Interrupt Pending   */
#define STAT_LUINT	(0x1<<1)        /* Line FIFO Under-Run Interrupt Pending        */
#define STAT_SINT	(0x1<<0)        /* System Error Interrupt Pending       */

/*******************************************************************************
 * MMIO BitBlt Module Registers
 *******************************************************************************
 */
#define REG_BB_CTRL	    0x1000  /* BitBlt module control register (write only) */
#define REG_BB_STAT	    0x1000  /* BitBlt module status register (read only) */

#define REG_BB_WINDOW	0x1004  /* Operation geometry */
#define REG_BB_SADDR	0x1008  /* Source start address */
#define REG_BB_DADDR	0x100c  /* Destination start address */
#define REG_BB_PITCH	0x1010  /* */
#define REG_BB_BG	    0x1014  /* Background color */
#define REG_BB_FG	    0x1018  /* Foreground color */

/* BitBlt status register bits */
#define BB_STAT_PROCESS	                (0x1<<31)       /* 1 - processing operation, 0 - idle */
#define BB_STAT_FULL	                (0x1<<30)       /* 1 - pipeline full */
#define BB_STAT_DMA	                    (0x1<<26)       /* DMA support */

#define BB_CTRL_CMD_MASK	            0xC0000000
#define BB_CTRL_CMD_START		        (0x1<<31)
#define BB_CTRL_CMD_ABORT		        (0x1<<30)

#define BB_CTRL_HALFWORD_TWISTER	    (0x1<<24)
#define BB_CTRL_BYTES_TWISTER		    (0x1<<23)
#define BB_CTRL_BITS_IN_BYTE_TWISTER	(0x1<<22)

#define BB_CTRL_DDMA_EN			        (0x1<<21)
#define BB_CTRL_SDMA_EN			        (0x1<<20)
#define BB_CTRL_SOFFS_MASK	            (0x7<<16)

/* Binary raster operations */
#define BB_CTRL_ROP_MASK		    0x0000F000

#define BB_CTRL_ROP_0			    (0x0<<12)       /* clear */
#define BB_CTRL_ROP_AND			    (0x1<<12)       /* and */
#define BB_CTRL_ROP_NOT_SRC_AND_DST	(0x2<<12)       /* andReverse */
#define BB_CTRL_ROP_DST			    (0x3<<12)       /* copy */
#define BB_CTRL_ROP_SRC_AND_NOT_DST	(0x4<<12)       /* andInverted */
#define BB_CTRL_ROP_SRC			    (0x5<<12)       /* noop */
#define BB_CTRL_ROP_XOR			    (0x6<<12)       /* xor */
#define BB_CTRL_ROP_OR			    (0x7<<12)       /* or */
#define BB_CTRL_ROP_NOR			    (0x8<<12)       /* nor */
#define BB_CTRL_ROP_NXOR		    (0x9<<12)       /* equiv */
#define BB_CTRL_ROP_NOT_SRC		    (0xa<<12)       /* invert */
#define BB_CTRL_ROP_NOT_SRC_OR_DST	(0xb<<12)       /* orReverse */
#define BB_CTRL_ROP_NOT_DST		    (0xc<<12)       /* copyInverted */
#define BB_CTRL_ROP_SRC_OR_NOT_DST	(0xd<<12)       /* orInverted */
#define BB_CTRL_ROP_NAND		    (0xe<<12)       /* nand */
#define BB_CTRL_ROP_1			    (0xf<<12)       /* set */

#define BB_CTRL_HDIR	            (0x1<<5)
#define BB_CTRL_VDIR	            (0x1<<6)

#define BB_CTRL_CE_EN		        (0x1<<0)
#define BB_CTRL_PAT_EN		        (0x1<<1)
#define BB_CTRL_SFILL_EN	        (0x1<<2)
#define BB_CTRL_TR_EN		        (0x1<<4)

#define BB_CTRL_SRC_MODE	        (0x1<<7)

#define BB_CTRL_TERM_00		        (0x0<<8)
#define BB_CTRL_TERM_01		        (0x1<<8)
#define BB_CTRL_TERM_10		        (0x2<<8)

#define BB_CTRL_BPP_8	            (0x0<<10)
#define BB_CTRL_BPP_16	            (0x1<<10)
#define BB_CTRL_BPP_24	            (0x2<<10)
#define BB_CTRL_BPP_32	            (0x3<<10)
#ifdef __sparc__
#define BB_CTRL_BPP_CD_8	        (BB_CTRL_BPP_8             )
#define BB_CTRL_BPP_CD_16	        (BB_CTRL_BPP_16 | 0x0800000)
#define BB_CTRL_BPP_CD_24	        (BB_CTRL_BPP_24 | 0x1800000)
#define BB_CTRL_BPP_CD_32	        (BB_CTRL_BPP_32 | 0x1800000)
#else  /* __sparc__ */
#define BB_CTRL_BPP_CD_8	        BB_CTRL_BPP_8
#define BB_CTRL_BPP_CD_16	        BB_CTRL_BPP_16
#define BB_CTRL_BPP_CD_24	        BB_CTRL_BPP_24
#define BB_CTRL_BPP_CD_32	        BB_CTRL_BPP_32
#endif /* __sparc__ */

typedef unsigned char uchar;

typedef struct {
    CARD32 sAddr;
    CARD32 dAddr;
    CARD32 pitch;
    CARD32 ctrl;
    CARD32 wind;
    CARD32 bg;
    CARD32 fg;
} MGARegsT;

/*******************************************************************************
 * Private data
 *******************************************************************************
 */

#define SCANLINE_NUM_BUFFERS	    64
#define PAGE_MASK                   (~(getpagesize() - 1))
#define MCSTMGA_SCANLINE_BUF_SIZE   8192
#define MCSTMGA_MEM_SIZE	        (0x800000 - (4*MCSTMGA_SCANLINE_BUF_SIZE))
#define MCSTMGA_CURSOR_SIZE	        0x4000

struct dmaMem {
    unsigned long physAddr;    /* dma addr */
    size_t size;
};

#ifdef USE_XAA
typedef struct {
    unsigned char *virtAddr;   /* buffer virt address */
    CARD32 physAddr;           /* buffer phys address */
    size_t size;               /* buffer size */
} xaaBufT;

typedef struct {
    CARD32 startAddr;          /* space address */
    size_t size;               /* space size */
} DMAListT;

#define	DMA_BUSY_LIST_SIZE	2
#endif /* USE_XAA */

typedef struct {
    int firmwareVersion;
    int firmwareDate;

    unsigned char *fbstart;     /* visible fb start address */
    unsigned char *fbmmio;      /* card registers base address */
    unsigned char *fbmem;       /* fb virtual base address */
    int fbmem_len;              /* fb mem size */

    int fboff;

    unsigned char *shadowmem;
    Bool shadowFB;

    void (*PointerMoved) (int index, int x, int y);
    CloseScreenProcPtr CloseScreen;
    EntityInfoPtr pEnt;
    int lineLength;

    /* DGA info */
    OptionInfoPtr Options;
    DGAModePtr pDGAMode;
    int nDGAMode;

    /* Cursor */
    xf86CursorInfoPtr CursorInfoRec;    /* cursor info pointer */
    unsigned long CursorImageOffset;
    Bool HWCursor;                      /* Use hardware cursor */

    /* Acceleration */
    Bool NoAccel;
    Bool useEXA;
    Bool NoDMA;
#ifdef USE_XAA
    XAAInfoRecPtr pXAAInfo;
    Bool useXAA;
#endif /* USE_XAA */
    ExaDriverPtr pEXAInfo;
    XF86ModReqInfo exaReq;
    XF86ModReqInfo xaaReq;

    /* Acceleration data */

    /* Solid fill parameters */
    int fillBytesPerPixel;
    CARD32 fillDstOffset;
    CARD16 fillPitch;
    int fillCardRop;            /* Video card ROP code */

    /* Copy parameters */
    int copyXDir, copyYDir;     /* direction for copy */
    int copyBytesPerPixel;      /* bytes per pixel */
    CARD32 copySrcOffset;
    CARD32 copyDstOffset;
    int copyCardRop;            /* Video card ROP code */
    int copySPitch;
    int copyDPitch;

    /* Color expand parameters */
    int expandTransparent;      /* Pixels coresponding to zeroes
                                 * in the bitmap left unmodified
                                 * by accelerator */
    int expandCardRop;          /* Video card ROP code */

    unsigned char *DmaChunkVBase;
    CARD32 ImageBytesPerPixel;
    CARD32 DmaChunkPhysBase;
    int DmaChunkUsedSize;
    int DmaChunkRange;
    MGARegsT MGA_regs;
    int BytesPerPixel;
    int imageX;
    int imageY;
    int imageW;

#ifdef MCSTMGA_DEBUG_ACCEL
#define DUMP_BLOCK_NUM 100
    CARD32 dumpBlock[DUMP_BLOCK_NUM][0x100];
    int dumpBlockIdx;
#endif

#ifdef USE_XAA
    DMAListT BusyDMAList[DMA_BUSY_LIST_SIZE];
    xaaBufT DirectBuff;
#endif /* USE_XAA */

} McstMgaRec, *McstMgaPtr;

#define McstMgaGetRec(p) ((McstMgaPtr)((p)->driverPrivate))

/*
 * Hardware cursor routines
 */
Bool McstMgaHWCursorInit(ScreenPtr pScreen);

/*
 * Hardware 2D acceleration routines
 */
void McstMgaAccelCleanup(ScreenPtr pScreen);
Bool McstMgaAccelInit(ScreenPtr pScreen);

static __inline__ CARD32
McstMgaFlip32(CARD32 d)
{
    return ((d & 0xff) << 24) | (((d >> 8) & 0xff) << 16) |
        (((d >> 16) & 0xff) << 8) | ((d >> 24) & 0xff);
}

static __inline__ CARD32
McstMgaMmio32Read(McstMgaPtr infoPtr, int reg)
{
#if defined(__sparc__) && defined(__sparc_v8__)
    /* Only for unpatched X.org builds where MMIO_... functions
     * depend on unsupported 0x88 asi */
    return McstMgaFlip32(*(volatile CARD32 *)(infoPtr->fbmmio + reg));
#else
    return MMIO_IN32(infoPtr->fbmmio, reg);
#endif
}

static __inline__ void
McstMgaMmio32Write(McstMgaPtr infoPtr, int reg, CARD32 value)
{
#ifdef MCSTMGA_DEBUG_ACCEL
    if (reg >= 0x1000) {
        infoPtr->dumpBlock[infoPtr->dumpBlockIdx][reg & 0xff] = value;
        if (reg == REG_BB_CTRL &&
            (value & BB_CTRL_CMD_START) == BB_CTRL_CMD_START) {
            infoPtr->dumpBlockIdx++;
            if (infoPtr->dumpBlockIdx >= DUMP_BLOCK_NUM) {
                infoPtr->dumpBlockIdx = 0;
            }
        }
    }
#endif

    if (reg == REG_BB_WINDOW) {
        if (value == infoPtr->MGA_regs.wind)
            return;
        else
            infoPtr->MGA_regs.wind = value;
    }

#if defined(__sparc__) && defined(__sparc_v8__)
    /* Only for unpatched X.org builds where MMIO_... functions
     * depend on unsupported 0x88 asi */
    *(volatile CARD32 *)(infoPtr->fbmmio + reg) = McstMgaFlip32(value);
#else
    MMIO_OUT32(infoPtr->fbmmio, reg, value);
#endif
}

#endif /* __MCSTMGA_H */
