/********************************************
*
* Copyright (c) 2022-2024 AO MCST
* Distributed under the terms of MIT License.
*
*********************************************/

#include <xorg-server.h>
#include "xf86.h"
#include "xf86_OSproc.h"

#include <sys/mman.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "fbdevhw.h"
#include "mcstmga.h"

#define USE_MCSTMGA_QUEUE 1
#define MCSTMGA_ERROR 0xFFFFFFFF

#define MCSTMGA_WATCHDOG_COUNTER 1000000
#define MCSTMGA_SHORT_WAIT_COUNTER 1000
/* Each request for a register value should last for approx. 300 clocks for E3M. Let's take 1 us for spinning... */
#define MCSTMGA_LARGE_DMA_TRANSFER_THRESHOLD 50000
/* Transfer speed for MCSTMGA is approx. 40 MB/s, so 50K should last just a bit more than 1us. */

#ifdef USE_XAA
extern void McstMgaXAA_SetupForScanlineImageWrite(ScrnInfoPtr pScrn, int rop,
                                                  unsigned int planemask,
                                                  int transColor, int bpp,
                                                  int depth);
extern void McstMgaXAA_SubsequentScanlineImageWriteRect(ScrnInfoPtr pScrn,
                                                        int x, int y, int w,
                                                        int h, int skipLeft);
extern void McstMgaXAA_SubsequentImageWriteScanline(ScrnInfoPtr pScrn,
                                                    int bufNo);
#endif /* USE_XAA */

/**
 * XAA to MGA/M-83 ROP codes translation.
 */
static const int ROPMap[16] = {
    BB_CTRL_ROP_0,                /* GXclear         */
    BB_CTRL_ROP_AND,              /* GXand           */
    BB_CTRL_ROP_SRC_AND_NOT_DST,  /* GXandReverse    */
    BB_CTRL_ROP_SRC,              /* GXcopy          */
    BB_CTRL_ROP_NOT_SRC_AND_DST,  /* GXandInverted   */
    BB_CTRL_ROP_DST,              /* GXnoop          */
    BB_CTRL_ROP_XOR,              /* GXxor           */
    BB_CTRL_ROP_OR,               /* GXor            */
    BB_CTRL_ROP_NOR,              /* GXnor           */
    BB_CTRL_ROP_NXOR,             /* GXequiv         */
    BB_CTRL_ROP_NOT_DST,          /* GXinvert        */
    BB_CTRL_ROP_SRC_OR_NOT_DST,   /* GXorReverse     */
    BB_CTRL_ROP_NOT_SRC,          /* GXcopyInverted  */
    BB_CTRL_ROP_NOT_SRC_OR_DST,   /* GXorInverted    */
    BB_CTRL_ROP_NAND,             /* GXnand          */
    BB_CTRL_ROP_1                 /* GXset           */
};

static int McstMgaWaitForFIFO(McstMgaPtr infoPtr, int as);

#ifdef USE_XAA
#ifdef MCSTMGA_DEBUG_ACCEL
static void
printDMABusyList(McstMgaPtr infoPtr)
{
    int i;

    ErrorF("Busy List:\n");
    for (i = 0; i < DMA_BUSY_LIST_SIZE; i++)
        ErrorF("%d | addr = 0x%8.8x  size = 0x%8.8x\n",
               i, infoPtr->BusyDMAList[i].startAddr,
               infoPtr->BusyDMAList[i].size);
}
#endif

static void
initDMAList(McstMgaPtr infoPtr)
{
    memset(infoPtr->BusyDMAList, 0, sizeof(DMAListT) * DMA_BUSY_LIST_SIZE);
}

static CARD32
getDMAAddr(McstMgaPtr infoPtr, size_t sz)
{
    /* Returns an offset to free space in DMA buffer, enough to hold sz bytes */
    CARD32 addr1;
    CARD32 addr;
    size_t size1;
    size_t size;

    if (sz >= MCSTMGA_MEM_SIZE) {
        return MCSTMGA_ERROR;
    }

    if (infoPtr->BusyDMAList[0].size == 0) {
        return (0);
    }

    addr = infoPtr->BusyDMAList[0].startAddr;
    size = infoPtr->BusyDMAList[0].size;

    if (infoPtr->BusyDMAList[1].size == 0) {
        if (sz <= addr)
            return (0);
        if ((MCSTMGA_MEM_SIZE - addr - size) >= sz)
            return (addr + size);
    }
    else {
        if (infoPtr->BusyDMAList[1].startAddr < addr) {
            addr1 = addr;
            size1 = size;
            addr = infoPtr->BusyDMAList[1].startAddr;
            size = infoPtr->BusyDMAList[1].size;
        }
        else {
            addr1 = infoPtr->BusyDMAList[1].startAddr;
            size1 = infoPtr->BusyDMAList[1].size;
        };
        if (sz <= addr)
            return (0);
        if ((addr1 - addr - size) >= sz)
            return (addr + size);
        if ((MCSTMGA_MEM_SIZE - addr1 - size1) >= sz)
            return (addr1 + size1);
    };

    McstMgaWaitForFIFO(infoPtr, 0);
    initDMAList(infoPtr);
    return (0);
}
#endif /* USE_XAA */

/**
 * This function is called to syncronize with the graphics engine and it waits
 * the graphic engine is idle. This is required before allowing direct access
 * to the framebuffer.
 * @param pScreenInfo	Screeen info pointer structure.
 * @return		none
 */
static void
McstMgaDebugPrintRegisters(McstMgaPtr infoPtr)
{
#ifdef MCSTMGA_DEBUG_ACCEL
    int i;
    for (i = 0; i <= infoPtr->dumpBlockIdx; i++) {
        xf86Msg(X_ERROR, "BitBlt registers block #%05d:\n"
                "\tREG_BB_CTRL 0x%08lx\n"
                "\tREG_BB_STAT 0x%08lx\n"
                "\tREG_BB_WINDOW 0x%08lx\n"
                "\tREG_BB_SADDR 0x%08lx\n"
                "\tREG_BB_DADDR 0x%08lx\n"
                "\tREG_BB_PITCH 0x%08lx\n"
                "\tREG_BB_BG 0x%08lx\n"
                "\tREG_BB_FG 0x%08lx\n",
                i,
                infoPtr->dumpBlock[i][REG_BB_CTRL & 0xff],
                infoPtr->dumpBlock[i][REG_BB_STAT & 0xff],
                infoPtr->dumpBlock[i][REG_BB_WINDOW & 0xff],
                infoPtr->dumpBlock[i][REG_BB_SADDR & 0xff],
                infoPtr->dumpBlock[i][REG_BB_DADDR & 0xff],
                infoPtr->dumpBlock[i][REG_BB_PITCH & 0xff],
                infoPtr->dumpBlock[i][REG_BB_BG & 0xff],
                infoPtr->dumpBlock[i][REG_BB_FG & 0xff]);
    }

    memset(infoPtr->dumpBlock, 0, sizeof(infoPtr->dumpBlock));
    infoPtr->dumpBlockIdx = 0;
    return;
#endif
}

static void
McstMgaWaitForIdle(ScrnInfoPtr pScrn)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int watchdog = MCSTMGA_WATCHDOG_COUNTER;

    /* wait until BitBlt module become idle */
#ifndef MCSTMGA_DEBUG_ACCEL
    while (McstMgaMmio32Read(infoPtr, REG_BB_STAT) & BB_STAT_PROCESS);
#else /* MCSTMGA_DEBUG_ACCEL */
    while (McstMgaMmio32Read(infoPtr, REG_BB_STAT) & BB_STAT_PROCESS) {
        if (watchdog-- > 0)
            continue;
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Waiting for idle...\n");
        {
            McstMgaDebugPrintRegisters(infoPtr);
        }
    }
#endif /* !MCSTMGA_DEBUG_ACCEL */
}

/**
 * Waits until next bitblt operation can be queued.
 * @param pScreenInfo	Screeen info pointer structure.
 * @param as		0 - wait BB_STAT_PROCESS, 1 - wait BB_STAT_FULL
 * @return		none
 */

/* Wait until all bits from andMaskToSpin are dropped */
static CARD32
McstMgaWaitForFIFOStatus(McstMgaPtr infoPtr, CARD32 andMaskToSpin,
                         int goDeepSleep)
{
    int countdown = MCSTMGA_SHORT_WAIT_COUNTER;
    int watchdog = MCSTMGA_WATCHDOG_COUNTER;
    CARD32 lastStatus = 0;

    /* First countdown cycle - only spin */
    while (countdown-- > 0) {
        if ((lastStatus =
             McstMgaMmio32Read(infoPtr, REG_BB_STAT)) & andMaskToSpin)
            continue;
        else
            return lastStatus;
    };
    watchdog -= MCSTMGA_SHORT_WAIT_COUNTER;
    /* Next countdown cycle - spin or sleep */
    do {
        if (goDeepSleep)
            usleep(1);
        else {
            countdown = MCSTMGA_SHORT_WAIT_COUNTER;
            while (countdown-- > 0) {
                if ((lastStatus =
                     McstMgaMmio32Read(infoPtr, REG_BB_STAT)) & andMaskToSpin)
                    continue;
                else
                    return lastStatus;
            };
        }
        if ((lastStatus =
             McstMgaMmio32Read(infoPtr, REG_BB_STAT)) & andMaskToSpin)
            watchdog -= MCSTMGA_SHORT_WAIT_COUNTER;
        else
            return lastStatus;
    } while (watchdog > 0);
    /* Timeout (watchdog) exceeded */
    return MCSTMGA_ERROR;
}

static int
McstMgaShallGoToDeepSleep(McstMgaPtr infoPtr) /* Reads from BusyDMAList */
{
#ifdef USE_XAA
    int totalOperationSize = 0;
    totalOperationSize += infoPtr->BusyDMAList[0].size;
    totalOperationSize += infoPtr->BusyDMAList[1].size;
    if (totalOperationSize > MCSTMGA_LARGE_DMA_TRANSFER_THRESHOLD)
        return 1;
    else
#endif /* USE_XAA */
        return 0;
}

static int
McstMgaWaitForFIFO(McstMgaPtr infoPtr, int as)
{
    int goDeepSleep = 0;
    int lastStatus = 0;

    /* wait until next bitblt operation can be queued */
    goDeepSleep = McstMgaShallGoToDeepSleep(infoPtr);
    if (as == 0) {
        McstMgaWaitForFIFOStatus(infoPtr, BB_STAT_PROCESS, goDeepSleep);
#ifdef USE_XAA
#if USE_MCSTMGA_QUEUE
        infoPtr->BusyDMAList[0].size = 0;
        infoPtr->BusyDMAList[1].size = 0;
#endif
#endif /* USE_XAA */
    }
    else {
        lastStatus =
            McstMgaWaitForFIFOStatus(infoPtr, BB_STAT_FULL, goDeepSleep);
        if ((lastStatus & (BB_STAT_FULL | BB_STAT_PROCESS)) != 0) {
#ifdef MCSTMGA_DEBUG_ACCEL
            if (lastStatus & BB_STAT_FULL) {
                /* Queue is full, obviously timeout. */
                McstMgaDebugPrintRegisters(infoPtr);
#ifdef USE_XAA
                printDMABusyList();
#endif /* USE_XAA */
                return MCSTMGA_ERROR;
            }
#endif
#ifdef USE_XAA
            /* ! BB_STAT_FULL && BB_STAT_PROCESS ==> (only) one request is active. */
            /* If both DMA slots are occupied, keep the most recent one */
            if (infoPtr->BusyDMAList[0].size > 0 &&
                infoPtr->BusyDMAList[1].size > 0) {
                infoPtr->BusyDMAList[0].size = infoPtr->BusyDMAList[1].size;
                infoPtr->BusyDMAList[0].startAddr =
                    infoPtr->BusyDMAList[1].startAddr;
                infoPtr->BusyDMAList[1].size = 0;
            }
#endif /* USE_XAA */
            /* One slot in queue is free */
            return 0;
        };
#ifdef USE_XAA
        /** !BB_STAT_FULL && !BB_STAT_PROCESS ==> Queue is empty
         * ==> both DMA slots are free
         */
        infoPtr->BusyDMAList[0].size = 0;
        infoPtr->BusyDMAList[1].size = 0;
#endif /* USE_XAA */
    }
    return 0;
}

/**
 * Set bitblt registers for solid fill operation.
 * Foreground color register have to be set before.
 */
static void
__SolidFill(ScrnInfoPtr pScrn, int x, int y, int width, int height)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    CARD32 ctrl = infoPtr->fillCardRop | BB_CTRL_CE_EN | BB_CTRL_SFILL_EN
        | ((infoPtr->fillBytesPerPixel - 1) & 0x3) << 10 | BB_CTRL_CMD_START;

    CARD32 dAddr = infoPtr->fillDstOffset
        + y * infoPtr->fillPitch
        + x * infoPtr->fillBytesPerPixel;

    CARD32 window = (height << 16) | (width * infoPtr->fillBytesPerPixel);

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, window);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl);
}

/*******************************************************************************
 * EXA Acceleration
 ******************************************************************************/

/**
 * This function is called to syncronize with the graphics engine and it waits
 * the graphic engine is idle. This is required before allowing direct access
 * to the framebuffer.
 * @param pScreenInfo	Screeen info pointer structure.
 * @return		none
 */
static void
McstMgaEXAWaitMarker(ScreenPtr pScreen, int marker)
{
    McstMgaWaitForIdle(xf86Screens[pScreen->myNum]);
}

/**
 * Setup for solid fill.
 * @param pPixmap Pixmap destination.
 * @param alu raster operation.
 * @param planemask mask for fill.
 * @param fg foreground color.
 *
 * Setup the card's engine for a solid fill operation into @a pPixmap.
 * @a alu specifies the raster op for the fill, @a planemask specifies an
 * optional mask, and @a fg specifies the foreground color for the fill.
 *
 * @return TRUE for if your driver can accelerate the operation with the
 * parameters given, or FALSE if EXA should fallback to manual rendering.
 */
static Bool
McstMgaEXAPrepareSolid(PixmapPtr pPixmap, int alu, Pixel planemask, Pixel fg)
{
    ScrnInfoPtr pScrn = xf86Screens[pPixmap->drawable.pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    /* Planemask not supported */
    if ((planemask & ((1 << pPixmap->drawable.depth) - 1)) !=
        (1 << pPixmap->drawable.depth) - 1)
        return FALSE;

    if (pPixmap->drawable.bitsPerPixel != 8 &&
        pPixmap->drawable.bitsPerPixel != 16 &&
        pPixmap->drawable.bitsPerPixel != 24 &&
        pPixmap->drawable.bitsPerPixel != 32)
        return FALSE;

    infoPtr->fillPitch = exaGetPixmapPitch(pPixmap);
    infoPtr->fillBytesPerPixel = pPixmap->drawable.bitsPerPixel >> 3;
    infoPtr->fillDstOffset = (CARD32) exaGetPixmapOffset(pPixmap);
    infoPtr->fillCardRop = ROPMap[alu];

    McstMgaWaitForFIFO(infoPtr, 0);
    McstMgaMmio32Write(infoPtr, REG_BB_FG, fg);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, infoPtr->fillPitch << 16 | 0);

    return TRUE;
}

/**
  * Solid fill operation.
  * @param pPixmap Pixmap destination.
  * @param x1 left coordinate.
  * @param y1 top coordinate.
  * @param x2 right coordinate.
  * @param y2 bottom coordinate.
  *
  * Perform the fill as specified by PrepareSolid, from x1,y1 to x2,y2 within
  * @a pPixmap.  Many devices just need the offset, pitch, start x,y, height
  * and width to do the fill; others may require you to calculate the
  * destination address including the offset plus the adjustment for the
  * pitch, x and y.  Check out a few driver implementations for ideas about
  * what might work best for your particular device.
  */
static void
McstMgaEXASolid(PixmapPtr pPixmap, int x1, int y1, int x2, int y2)
{
    ScrnInfoPtr pScrn = xf86Screens[pPixmap->drawable.pScreen->myNum];

    __SolidFill(pScrn, x1, y1, x2 - x1, y2 - y1);
}

/**
 * Finish a solid fill.
 * @param pPixmap Pixmap to finish.
 *
 * Finish the solid fills done in the preceding Solid calls.
 */
static void
McstMgaEXADoneSolid(PixmapPtr pPixmap)
{
    /* Do nothing */
}

/**
 * PrepareCopy - setup a copy operation
 * @param pSrcPixmap source Pixmap
 * @param pDstPixmap destination Pixmap
 * @param xDir x direction for the copy
 * @param yDir y direction for the copy
 * @param alu raster operation
 * @param planemask optional planemask for the copy
 *
 * Copy @a pSrcPixmap to @a pDstPixmap in the x and y direction specified,
 * with the raster operation @a alu.  @planemask specifies an optional
 * color plane mask for the copy.  Note that @a pSrcPixmap isn't available in
 * the subsequent Copy routine, so you may have to store some of its values
 * elsewhere for later use.
 *
 * You can add additional fields to your driver record structure to store
 * state needed by this routine, if necessary, for instance calculating the
 * framebuffer offset in PrepareCopy and then using it in the subsequent
 * Copy call.
 *
 * Return TRUE if driver can accelerate the copy operation with the
 * parameters given, or FALSE if EXA should fallback to manual rendering.
 */
static Bool
McstMgaEXAPrepareCopy(PixmapPtr pSrcPixmap, PixmapPtr pDstPixmap,
                      int xDir, int yDir, int alu, Pixel planemask)
{
    ScrnInfoPtr pScrn = xf86Screens[pSrcPixmap->drawable.pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    /* Planemask not supported */
    if ((planemask & ((1 << pSrcPixmap->drawable.depth) - 1)) !=
        (1 << pSrcPixmap->drawable.depth) - 1)
        return FALSE;

    /* XXX Different bitsPerPixel not supported */
    if (pSrcPixmap->drawable.bitsPerPixel != pDstPixmap->drawable.bitsPerPixel)
        return FALSE;

    infoPtr->copyXDir = xDir;
    infoPtr->copyYDir = yDir;
    infoPtr->copyBytesPerPixel = pSrcPixmap->drawable.bitsPerPixel >> 3;
    infoPtr->copySPitch = exaGetPixmapPitch(pSrcPixmap);
    infoPtr->copyDPitch = exaGetPixmapPitch(pDstPixmap);
    infoPtr->copySrcOffset = (CARD32) exaGetPixmapOffset(pSrcPixmap);
    infoPtr->copyDstOffset = (CARD32) exaGetPixmapOffset(pDstPixmap);
    infoPtr->copyCardRop = ROPMap[alu];

    return TRUE;
}

/**
 * Copy - perform a copy between two pixmaps
 * @param pDstPixmap destination Pixmap
 * @param srcX source X coordinate
 * @param srcY source Y coordinate
 * @param dstX destination X coordinate
 * @param dstY destination Y coordinate
 * @param width copy width
 * @param height copy height
 *
 * Perform the copy setup by the previous PrepareCopy call, from
 * (@srcX,@srcY) in the source pixmap to (@dstX,@dstY) in pDstPixmap using
 * @width and @height to determine the quantity of the copy.
 */
static void
McstMgaEXACopy(PixmapPtr pDstPixmap, int srcX, int srcY, int dstX, int dstY,
               int width, int height)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPixmap->drawable.pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 ctrl = infoPtr->copyCardRop; /* Control register */
    CARD32 sAddr, dAddr;
    CARD32 align = 0;

    if (infoPtr->copyXDir < 0) {
        ctrl |= BB_CTRL_HDIR;
        srcX += width - 1;
        dstX += width - 1;
        align = infoPtr->copyBytesPerPixel - 1;
    }

    if (infoPtr->copyYDir < 0) {
        ctrl |= BB_CTRL_VDIR;
        srcY += height - 1;
        dstY += height - 1;
    }

    sAddr = infoPtr->copySrcOffset
        + srcY * infoPtr->copySPitch
        + srcX * infoPtr->copyBytesPerPixel + align;
    dAddr = infoPtr->copyDstOffset
        + dstY * infoPtr->copyDPitch
        + dstX * infoPtr->copyBytesPerPixel + align;

    McstMgaWaitForFIFO(infoPtr, 0);

    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW,
                       (height << 16) | (width * infoPtr->copyBytesPerPixel));
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, sAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH,
                       infoPtr->copyDPitch << 16 | infoPtr->copySPitch);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl | BB_CTRL_CMD_START);
}

/**
 * DoneCopy - finish a copy operation
 * @param pDstPixmap Pixmap to complete
 *
 * Tear down the copy operation for @pDstPixmap, if necessary.
 */
static void
McstMgaEXADoneCopy(PixmapPtr pDstPixmap)
{
    /* Do nothing */
}

/**
 * UploadToScreen - loads a rectangle of data from @src into @pDstPixmap.
 *
 * @param pDstPixmap - destination pixmap
 * @param x - destination x coordinate
 * @param y - destination y coordinate
 * @param width - width of the rectangle to be copied
 * @param height - height of the rectangle to be copied
 * @param src - pointer to the beginning of the source data
 * @param srcPitch - pitch (in bytes) of the lines of source data
 *
 * Returns
 *  TRUE if the driver successfully uploaded the data.
 *	FALSE indicates that EXA should fall back to doing the upload in software.
 */
static Bool
McstMgaEXAUploadToScreen(PixmapPtr pDstPixmap, int x, int y,
                         int width, int height, char *src, int srcPitch)
{
    ScrnInfoPtr pScrn = xf86Screens[pDstPixmap->drawable.pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 dstPitch;
    CARD32 dstIdx;
    CARD32 srcIdx;
    CARD32 bbCtrl;
    int available;
    CARD32 bppOp;
    CARD32 lines;
    CARD32 bpp;
    CARD8 *dst;

    bpp = pDstPixmap->drawable.bitsPerPixel;
    dstPitch = exaGetPixmapPitch(pDstPixmap);

    dst = infoPtr->DmaChunkVBase;       /* Virtual addr */
    srcIdx = infoPtr->DmaChunkPhysBase;        /* Physical addr */

    dstIdx = (y * dstPitch) + (x * (bpp >> 3));

    lines = height;

    switch (bpp) {
    case 32:
        bppOp = BB_CTRL_BPP_CD_32;
        break;
    case 24:
        bppOp = BB_CTRL_BPP_CD_24;
        break;
    case 16:
        bppOp = BB_CTRL_BPP_CD_16;
        break;
    case 8:
        bppOp = BB_CTRL_BPP_CD_8;
        break;
    default:
        return FALSE;
    };

    bbCtrl = (ROPMap[3] | bppOp | BB_CTRL_SDMA_EN);

    width = width * (bpp >> 3);

    available = infoPtr->DmaChunkRange - infoPtr->DmaChunkUsedSize;

    /*
     * Reduce the maximum available DMA buffer size in order to avoid
     * problems with speculative reading from IOMMU by video card.
     *
     */
    available -= srcPitch;

    if (available > width * ((signed int)lines)) {
        McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    }
    else {
        /*
         * DMA buffer is full. Waiting for FIFO processing is complete and start
         * filling it from the beginning.
         */
        McstMgaWaitForFIFO(infoPtr, 0);
        infoPtr->DmaChunkUsedSize = 0;
    }

    /*
     * Coping from high memory address (system memory) to low memory address
     * (system memory that suitable for dma 32). The object in low memory must be
     * filled uninterruptedly for the card to work in BB_CTRL_SRC_MODE
     */

    if (srcPitch == width) {
        memcpy(dst + infoPtr->DmaChunkUsedSize, src, width * lines);
    }
    else {
#pragma loop count(1000)
        while (lines--) {
            memcpy(dst + infoPtr->DmaChunkUsedSize, src, width);
            src += srcPitch;
            dst += width;
        }
    }

    srcIdx += infoPtr->DmaChunkUsedSize;
    infoPtr->DmaChunkUsedSize += width * height;

    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, ((height << 16) | width));
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, srcIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dstIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, ((dstPitch << 16) | width));
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, bbCtrl | BB_CTRL_CMD_START);

    return TRUE;
}

/**
 * DownloadFromScreen - loads a rectangle of data from @pSrcPixmap into @dst.
 *
 * @parm pSrcPixmap - source pixmap
 * @parm x - source x coordinate
 * @parm y - source y coordinate
 * @parm width - width of the rectangle to be copied
 * @parm height - height of the rectangle to be copied
 * @parm dst - pointer to the beginning of the destination data
 * @parm dstPitch - pitch (in bytes) of the lines of destination data
 *
 * Returns
 *	TRUE if the driver successfully downloaded the data.
 *	FALSE indicates that EXA should fall back to doing the download in software.
 */
static Bool
McstMgaEXADownloadFromScreen(PixmapPtr pSrcPixmap, int x, int y,
                             int width, int height, char *dst, int dstPitch)
{
    ScrnInfoPtr pScrn = xf86Screens[pSrcPixmap->drawable.pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 srcPitch;
    int available;
    CARD32 dstIdx;
    CARD32 srcIdx;
    CARD32 bbCtrl;
    CARD32 bppOp;
    CARD32 lines;
    CARD32 bpp;
    CARD8 *src;
    CARD32 i;

    bpp = pSrcPixmap->drawable.bitsPerPixel;
    srcPitch = exaGetPixmapPitch(pSrcPixmap);

    switch (bpp) {
    case 32:
        bppOp = BB_CTRL_BPP_CD_32;
        break;
    case 24:
        bppOp = BB_CTRL_BPP_CD_24;
        break;
    case 16:
        bppOp = BB_CTRL_BPP_CD_16;
        break;
    case 8:
        bppOp = BB_CTRL_BPP_CD_8;
        break;
    default:
        return FALSE; /* fall back to software EXA proc */
    };

    bbCtrl = (BB_CTRL_ROP_SRC | bppOp | BB_CTRL_DDMA_EN);

    width = width * (bpp >> 3);

    available = infoPtr->DmaChunkRange - infoPtr->DmaChunkUsedSize;

    if (available > width * height) {
        McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    }
    else {
        /*
         * DMA buffer is full. Waiting for FIFO processing is complete and start
         * filling it from the beginning.
         */
        McstMgaWaitForFIFO(infoPtr, 0);
        infoPtr->DmaChunkUsedSize = 0;
    }

    /* source x, source y and source bpp */
    srcIdx = (y * srcPitch) + (x * (bpp >> 3));

    /* Physical addr of the DMA buffer start + used size */
    dstIdx = infoPtr->DmaChunkPhysBase + infoPtr->DmaChunkUsedSize;

    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, (height << 16) | width);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, srcIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dstIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, (dstPitch << 16) | srcPitch);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, bbCtrl | BB_CTRL_CMD_START);

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);

    /*
     * Дожидаемся окончания DMA пересылок (записываем первый байт обратно в карту:
     * так будет гарантировано, что всё считано из видеопамяти в память процессора.)
     */

    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, (1 << 16) | 1);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, srcIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dstIdx);
    for (i = 0; i < 8; i++) {
        McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
        McstMgaMmio32Write(infoPtr, REG_BB_CTRL, bbCtrl | BB_CTRL_CMD_START);
    }

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);

    src = infoPtr->DmaChunkVBase + infoPtr->DmaChunkUsedSize;

    if (srcPitch == width) {
        memcpy(dst, src, width * height);
    }
    else {
        lines = height;
#pragma loop count(1000)
        while (lines--) {
            memcpy(dst, src, width);
            src += width;
            dst += dstPitch;
        }
    }

    infoPtr->DmaChunkUsedSize += width * height;

    return TRUE;
}

/**
 * EXA Acceleration init
 * @param pScreenInfo	Screen info pointer structure.
 * @return		success flag
 */
static Bool
McstMgaEXAInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int fd = fbdevHWGetFD(pScrn);
    ExaDriverPtr pEXAInfo;
    struct dmaMem dmem;

    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "EXAInit start execution\n");

    pEXAInfo = calloc(1, sizeof(ExaDriverRec));
    if (!pEXAInfo) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "failed to allocate EXA record\n");
        return FALSE;
    }

    pEXAInfo->needsSync = TRUE;
    pEXAInfo->memoryBase = infoPtr->fbmem;
    pEXAInfo->memorySize = infoPtr->fbmem_len - MCSTMGA_CURSOR_SIZE;
    pEXAInfo->offScreenBase = pScrn->virtualY * pScrn->virtualX
        * (pScrn->bitsPerPixel >> 3);

    if (pEXAInfo->memorySize > pEXAInfo->offScreenBase) {
        pEXAInfo->flags |= EXA_OFFSCREEN_PIXMAPS;
    }
    else {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Not enough video RAM for offscreen memory manager. Xv disabled\n");
    }

    pEXAInfo->flags |= EXA_TWO_BITBLT_DIRECTIONS;

    pEXAInfo->pixmapOffsetAlign = 0x1000;
    pEXAInfo->pixmapPitchAlign =
        pScrn->displayWidth * (pScrn->bitsPerPixel >> 3);
    pEXAInfo->pixmapOffsetAlign = 16;

    pEXAInfo->exa_major = EXA_VERSION_MAJOR;
    pEXAInfo->exa_minor = EXA_VERSION_MINOR;

    pEXAInfo->maxX = (65536 / pScrn->bitsPerPixel) - 1;
    pEXAInfo->maxY = (1 << 16) - 1;

    pEXAInfo->WaitMarker = McstMgaEXAWaitMarker;

    pEXAInfo->PrepareSolid = McstMgaEXAPrepareSolid;
    pEXAInfo->Solid = McstMgaEXASolid;
    pEXAInfo->DoneSolid = McstMgaEXADoneSolid;

    pEXAInfo->PrepareCopy = McstMgaEXAPrepareCopy;
    pEXAInfo->Copy = McstMgaEXACopy;
    pEXAInfo->DoneCopy = McstMgaEXADoneCopy;

    /* Dma routines */
    do {
        if (infoPtr->NoDMA == TRUE) {
            xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                       "DMA disabled, skip EXA_UploadToScreen() init\n");
            break;
        }
        ioctl(fd, FBIOFREE_ALL_DMA_MEMS, NULL);
        dmem.size = 0x800000;   /* 8 MB */
        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                   "McstMgaEXAInit: alllocating a chunk "
                   "of system memory to be used for dme needs: 0x%lx bytes\n",
                   dmem.size);
        if (0 != ioctl(fd, FBIOALLOC_DMA_MEM, (void *)(&dmem))) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "FBIOALLOC_DMA_MEM: %s\n", strerror(errno));
            break;
        }
        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                   "McstMgaEXAInit: FBIOALLOC_DMA_MEM "
                   "0x%lx bytes of dma mem at physAddr 0x%lx allocated\n",
                   dmem.size, dmem.physAddr);

        infoPtr->DmaChunkVBase = mmap(NULL, dmem.size, PROT_READ | PROT_WRITE,
                                      MAP_SHARED, fd,
#ifdef __sparc__
                                      dmem.physAddr);
#else
                                      dmem.physAddr + 0x80000000);
#endif
        if (-1 == (long)infoPtr->DmaChunkVBase) {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "McstMgaEXAInit: mmaping dma buffer failed: %s\n",
                       strerror(errno));
            ioctl(fd, FBIOFREE_ALL_DMA_MEMS, NULL);
            break;
        }

        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "McstMgaEXAInit: mapping "
                   "dma buffer at virtAddr 0x%lx with size 0x%x succeded\n",
                   (unsigned long)infoPtr->DmaChunkVBase, dmem.size);

        infoPtr->DmaChunkRange = dmem.size;
        infoPtr->DmaChunkPhysBase = dmem.physAddr;
        infoPtr->DmaChunkUsedSize = 0;

        pEXAInfo->UploadToScreen = McstMgaEXAUploadToScreen;
        pEXAInfo->DownloadFromScreen = McstMgaEXADownloadFromScreen;
    } while (0);

    if (!exaDriverInit(pScreen, pEXAInfo)) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "failed to init EXA\n");
        free(pEXAInfo);
        return FALSE;
    }

    infoPtr->pEXAInfo = pEXAInfo;
    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
               "EXAInit finish execution successfuly\n");
    return TRUE;
}

/**
 * EXA Acceleration cleanup
 * @param pScreenInfo	Screeen info pointer structure.
 * @return		none
 */
static void
McstMgaEXA_Cleanup(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    exaDriverFini(pScreen);

    if (infoPtr->pEXAInfo) {
        free(infoPtr->pEXAInfo);
        infoPtr->pEXAInfo = NULL;
    }
}

/*******************************************************************************
 * XAA Acceleration
 ******************************************************************************/
#ifdef USE_XAA

#define USE_XAA_DMA

#ifdef USE_XAA_DMA
/**
 * XAA_GetDMAMemory
 *
 * Description:	This routine is used to get DMA buffers for XAA
 *
 * Parameters.
 *	pScreenInfo	Screen handler pointer having screen information.
 *
 * Returns:		-1 - error. Can not allocate dma buffers.
 *			0  - ok
 * Comments:		none
 *
 */
static int
XAA_GetDMAMemory(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int fd = fbdevHWGetFD(pScrn);
    unsigned char *addr;
    struct dmaMem dmem;
    size_t bsz;
    size_t sz;
    int Bpp; /* bytes per pixel */

    /* DMA support */

    Bpp = pScrn->bitsPerPixel >> 3;
    sz = pScrn->virtualX * Bpp; /* line size (bytes) */
    bsz = (sz + ~PAGE_MASK) & PAGE_MASK; /* buffer size (bytes) */
    infoPtr->DirectBuff.size = MCSTMGA_MEM_SIZE;
    dmem.size = MCSTMGA_MEM_SIZE + (4 * MCSTMGA_SCANLINE_BUF_SIZE);

    if (ioctl(fd, FBIOALLOC_DMA_MEM, (void *)(&dmem)) != 0) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "FBIOALLOC_DMA_MEM: %s\n",
                   strerror(errno));
        return (-1);
    };
    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "FBIOALLOC_DMA_MEM "
               "0x%lx bytes of dma mem at physAddr 0x%lx allocated\n",
               dmem.size, dmem.physAddr);
    ErrorF("XAA_GetDMAMemory: size = 0x%lx  phys addr = 0x%lx\n",
           dmem.size, dmem.physAddr);
    infoPtr->DirectBuff.physAddr = dmem.physAddr;
    addr =
        (unsigned char *)mmap(NULL, dmem.size, PROT_READ | PROT_WRITE,
                               MAP_SHARED, fd,
#ifdef __sparc__
                               dmem.physAddr);
#else
                               dmem.physAddr + 0x80000000);
#endif
    ErrorF("XAA_GetDMAMemory: virt addr = 0x%lx\n", addr);
    if ((long) addr == -1) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "mmaping dma buffer failed: %s\n",
                   strerror(errno));
        return (-1);
    };
    infoPtr->DirectBuff.virtAddr = addr;
    initDMAList(infoPtr);
    return (0);
}
#endif /* USE_XAA_DMA */

/**
 * McstMgaXAASetupForScreenToScreenCopy
 *
 * Description:		This function is used to set up the planemask and raster op
 *			for future Bliting
 * functionality.
 * @param pScreenInfo	Screen handler pointer having screen information.
 * @param xDir		This is set based on rop data.
 * @param yDir		This is set based on rop data.
 * @param rop		sets the raster operation
 * @param planemask	Specifies the value of masking from rop data
 * @param transColor
 * @return		none
 */
static void
McstMgaXAASetupForScreenToScreenCopy(ScrnInfoPtr pScrn, int xDir, int yDir,
                                     int rop, unsigned int planemask, int transColor)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 ctrl;

    ctrl = ROPMap[rop];
    if (xDir == -1)
        ctrl |= BB_CTRL_HDIR;
    if (yDir == -1)
        ctrl |= BB_CTRL_VDIR;
    infoPtr->MGA_regs.ctrl = ctrl;
}

/**
 * McstMgaXAASubsequentScreenToScreenCopy
 *
 * Description:		This function is called to perform a screen to screen BLT using
 * 			the previously specified planemask,raster operation and
 *			transparency flag.
 *
 * The patterns specified is ignored inside the function.
 * Sample application uses (non-transparent):
 * @li Moving windows.
 * @li x11perf -scroll500 	(scroll tests).
 * @li x11perf -copywinwin500	(copy from window to window).
 *
 * No application found using transparency.
 *
 * @param pScreenInfo	Screen handler pointer having screen information.
 * @param x1		x-coordinates of the source window
 * @param y1    	y-coordinates of the source window
 * @param x2		x-coordinates of the destination window
 * @param y2    	y-coordinates of the destination window
 * @param w		Specifies width of the window to be copied
 * @param h     	Height of the window to be copied.
 * @return		none
 */
static void
McstMgaXAASubsequentScreenToScreenCopy(ScrnInfoPtr pScrn, int xSrc, int ySrc,
                                       int xDst, int yDst, int w, int h)
{
    int pitch = pScrn->displayWidth * bytesPerPixel;
    int bytesPerPixel = pScrn->bitsPerPixel >> 3;
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 sAddr;
    CARD32 dAddr;
    CARD32 ctrl;
    int al = 0;

    ctrl = infoPtr->MGA_regs.ctrl | BB_CTRL_CMD_START;

    if ((ctrl & BB_CTRL_HDIR) == BB_CTRL_HDIR) {
        xSrc = xSrc + w - 1;
        xDst = xDst + w - 1;
        al = bytesPerPixel - 1;
    }
    if ((ctrl & BB_CTRL_VDIR) == BB_CTRL_VDIR) {
        ySrc = ySrc + h - 1;
        yDst = yDst + h - 1;
    }

    sAddr = (ySrc * pScrn->displayWidth + xSrc) * bytesPerPixel + al;
    dAddr = (yDst * pScrn->displayWidth + xDst) * bytesPerPixel + al;

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, (h << 16) | (w * bytesPerPixel));
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, sAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, pitch << 16 | pitch);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl);
}

/**
 * McstMgaXAASetupForScreenToScreenColorExpandFill
 *
 * Description:		This routine is called to setup the following
 *			parameters used by the next call
 *			McstMgaXAASubsequentScreenToScreenColorExpandFill.
 *
 * Parameters:
 * 	pScrn		Screen handler pointer having screen information.
 * 	fb		Specifies foreground color being needed for color expansion
 * 			mode (Each 1 of bitmap forms fg colored pixel).
 *	bg		Specifies background color being needed for color expansion
 * 			mode (Each 0 of bitmap forms bg colored pixel).
 * 	rop		Specifies the raster operation value.
 * 	planemask	Specifies the masking value based rop srcdata.
 *
 * Returns:		none
 *
 * Comments:		none
 *
 *----------------------------------------------------------------------------
 */

void
McstMgaXAASetupForScreenToScreenColorExpandFill(ScrnInfoPtr pScrn, int fg,
                                                int bg, int rop,
                                                unsigned int planemask)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int bpp = pScrn->bitsPerPixel;
    CARD32 bbCtrl;
    int bppOp;

    switch (pScrn->bitsPerPixel) {
    case 32:
        bppOp = BB_CTRL_BPP_32;
        break;
    case 24:
        bppOp = BB_CTRL_BPP_24;
        break;
    case 16:
        bppOp = BB_CTRL_BPP_16;
        break;
    case 8:
        bppOp = BB_CTRL_BPP_8;
        break;
    default:
        return;
    };
    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);

    McstMgaMmio32Write(infoPtr, REG_BB_BG, bg);
    McstMgaMmio32Write(infoPtr, REG_BB_FG, fg);

    bbCtrl = BB_CTRL_CE_EN | ROPMap[rop] | bppOp;
    if (bg == -1)
        bbCtrl |= BB_CTRL_TR_EN;
    infoPtr->MGA_regs.ctrl = bbCtrl;
}

/**
 * McstMgaXAASubsequentScreenToScreenColorExpandFill
 *
 * Description:		This routine is called to start color expantion
 *			copy operation from offscreen framebuffer memory to
 *			it's visible part. The function is subsequent after
 *			McstMgaXAASetupForScreenToScreenColorExpandFill.
 *
 * Parameters:
 * 	pScrn		Screen handler pointer having screen information.
 * 	x		Specifies horizontal coordinate of destination window.
 *	y		Specifies vertical coordinate of destination window.
 *			x and y coordinates are in units of pixels
 *			Note that coordinates of destination window are the same
 *			of the top left pixel. It is right only for that function while
 *			in other cases it may be requiered to allow given parameters
 *			determining for example the direction of object rendering
 *			(left --> write, write <-- left; top->down; down<-top)
 * 	w		Object width specified in pixels.
 * 	h		Object height specified in pixels.
 *	srcX		Specifies horizontal source component of the object to be rendered
 *	srcY		Specifies vertical source component of the object to be rendered
 *			Note that object source stored in offscreen framebuffer memory is
 *			expressed as bitmap. It is the same as visible object (destination)
 *			but only taken smaller space due to that each bit of that one
 *			determines either fg pixel of destination when in 1 state or bg pixel
 *			if being zeroed.
 *	offset		Determines offset to the first bit of given by pair (srcX:srcY) byte
 *			Offset is allowed for each line.
 * Returns:		none
 *
 * Comments:		none
 *
 *----------------------------------------------------------------------------
 */

void
McstMgaXAASubsequentScreenToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                  int x, int y, int w, int h,
                                                  int srcX, int srcY,
                                                  int offset)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 bbCtrl = infoPtr->MGA_regs.ctrl;
    int pitch = pScrn->displayWidth * bpp;
    int bpp = pScrn->bitsPerPixel >> 3;
    int dstIdx = 0;
    int srcIdx = 0;

    w = w * bpp;                /* width in Bytes */

    dstIdx = y * pitch + x * bpp;      /* Bytes */
    srcIdx = srcY * pitch + srcX * bpp;        /* Bytes */

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);

    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, (h << 16) | w);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, srcIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, dstIdx);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, pitch << 16 | pitch);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, bbCtrl |
                       ((offset << 16) & BB_CTRL_SOFFS_MASK) |
                       BB_CTRL_CMD_START);
}

/**
 * McstMgaXAASetupForScanlineCPUToScreenColorExpandFill
 *
 * Description:		This routine is called to setup the following
 *			parameters used by the next call
 *			McstMgaXAASubsequentScanlineCPUToScreenColorExpandFill.
 *
 * Parameters:
 * 	pScrn		Screen handler pointer having screen information.
 * 	fb		Specifies foreground color being needed for color expansion
 * 			mode (Each 1 of bitmap forms fg colored pixel).
 *	bg		Specifies background color being needed for color expansion
 * 			mode (Each 0 of bitmap forms bg colored pixel).
 * 	rop		Specifies the raster operation value.
 * 	planemask	Specifies the masking value based rop srcdata.
 *
 * Returns:		none
 *
 * Comments:		none
 *
 *----------------------------------------------------------------------------
 */
void
McstMgaXAASetupForScanlineCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                     int fg, int bg, int rop,
                                                     unsigned int planemask)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int pitch = pScrn->displayWidth * bpp;
    int bpp = pScrn->bitsPerPixel;
    CARD32 bbCtrl;
    int bppOp;

    switch (pScrn->bitsPerPixel) {
    case 32:
        bppOp = BB_CTRL_BPP_32;
        break;
    case 24:
        bppOp = BB_CTRL_BPP_24;
        break;
    case 16:
        bppOp = BB_CTRL_BPP_16;
        break;
    case 8:
        bppOp = BB_CTRL_BPP_8;
        break;
    default:
        return;
    };

    bbCtrl = BB_CTRL_CE_EN | BB_CTRL_SDMA_EN | bppOp | BB_CTRL_CMD_START;
#if ByteOrder == X_BIG_ENDIAN
    /**
	 * Чтобы не выставлять для sparc флаг BIT_ORDER_IN_BYTE_MSBFIRST
	 * (по которому биты в байтах будут переворачиваться программно),
	 * используем возможность в самом BitBlt по вращению бит.
	 * Это на несколько процентов ускорит вывод текста на экран.
	 */
    bbCtrl |= BB_CTRL_BITS_IN_BYTE_TWISTER;
#endif
    if (bg == -1)
        bbCtrl |= BB_CTRL_TR_EN;
    infoPtr->MGA_regs.ctrl = bbCtrl | ROPMap[rop];

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    McstMgaMmio32Write(infoPtr, REG_BB_BG, bg);
    McstMgaMmio32Write(infoPtr, REG_BB_FG, fg);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, pitch << 16 | 0);
}

/**
 * McstMgaXAASubsequentScanlineCPUToScreenColorExpandFill
 *
 * Description:		This routine is called after each scan line to request DMA buffer for
 *			color expantion	copy operation from system memory.
 *			The function is subsequent after
 *			McstMgaXAASetupForScanlineCPUToScreenColorExpandFill.
 *
 * Parameters:
 * 	pScrn		Screen handler pointer having screen information.
 * 	x		Specifies horizontal coordinate of destination window.
 *	y		Specifies vertical coordinate of destination window.
 *			x and y coordinates are in units of pixels
 *			Note that coordinates of destination window are the same
 *			of the top left pixel. It is right only for that function while
 *			in other cases it may be requiered to allow given parameters
 *			determining for example the direction of object rendering
 *			(left --> write, write <-- left; top->down; down<-top)
 * 	w		Object width specified in pixels.
 * 	h		Object height specified in pixels.
 *	skipLeft	If the skipLeft is non-zero (and LEFT_EDGE_CLIPPING has been
 *                      enabled), then the accelerator _should_not_ render skipLeft
 *                      pixels on the leftmost edge of the rectangle.
 *
 * Returns:		none
 *
 * Comments:		none
 *
 *----------------------------------------------------------------------------
 */
void
McstMgaXAASubsequentScanlineCPUToScreenColorExpandFill(ScrnInfoPtr pScrn,
                                                       int x, int y,
                                                       int w, int h,
                                                       int skipLeft)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int bpp = pScrn->bitsPerPixel >> 3;
    int pitch = pScrn->displayWidth * bpp;
    CARD32 window;

    window = (1 << 16) | (w * bpp);     /* h=1 */
    infoPtr->MGA_regs.dAddr = y * pitch + x * bpp;
    McstMgaWaitForFIFO(infoPtr, 0);
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, window);
}

static unsigned char *_scanline_buf[4];

/**
 * McstMgaXAASubsequentColorExpandScanlinee
 *
 * Description:		This routine is called after each scanline for draw it.
 *			McstMgaXAASubsequentScanlineCPUToScreenColorExpandFill.
 *
 * Parameters:
 * 	pScrn		Screen handler pointer having screen information.
 * 	bufNo		Number of buffer to draw.
 *
 * Returns:		none
 *
 * Comments:		none
 *
 * Sample application uses:
 *   - x11perf -ftext (Char in 80-char line).
 *----------------------------------------------------------------------------
 */
void
McstMgaXAASubsequentColorExpandScanlinee(ScrnInfoPtr pScrn, int bufNo)
{
    CARD32 sAddr = infoPtr->DirectBuff.physAddr + MCSTMGA_MEM_SIZE
        + (bufNo * MCSTMGA_SCANLINE_BUF_SIZE);
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 bbCtrl = infoPtr->MGA_regs.ctrl;
    int pitch = pScrn->displayWidth * bpp;
    int bpp = pScrn->bitsPerPixel >> 3;

    McstMgaWaitForFIFO(infoPtr, (bufNo == 3) ? 0 : USE_MCSTMGA_QUEUE);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, sAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, infoPtr->MGA_regs.dAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, bbCtrl);
    infoPtr->MGA_regs.dAddr += pitch;   /* Переходим на следующую строку */
}

/**
 * McstMgaSetupForSolidFill
 *
 * Description:		This routine is called to setup the solid pattern
 * 			color for future rectangular fills or vectors.
 *
 * Parameters:
 * 	pScrn		Screen handler pointer having screen information.
 * 	color		Specifies the color to be filled up in defined area.
 * 	rop		Specifies the raster operation value.
 * 	planemask	Specifies the masking value based rop srcdata.
 *
 * Returns:		none
 *
 * Comments:		none
 *
 *----------------------------------------------------------------------------
 */
static void
McstMgaXAASetupForSolidFill(ScrnInfoPtr pScrn, int color, int rop,
                             unsigned int planemask)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    infoPtr->fillBytesPerPixel = pScrn->bitsPerPixel >> 3;
    infoPtr->fillPitch = pScrn->displayWidth * infoPtr->fillBytesPerPixel;
    infoPtr->fillDstOffset = 0;
    infoPtr->fillCardRop = ROPMap[rop];

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    McstMgaMmio32Write(infoPtr, REG_BB_FG, color);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, infoPtr->fillPitch << 16 | 0);
}

 /**
  * McstMgaSubsequentSolidFillRect --
  *
  * Description:	This routine is used to fill the rectangle of
  * 			previously specified  solid pattern.
  *
  * Parameters.
  *	pScreenInfo	Screen handler pointer having screen information.
  *	x and y		Specifies the x and y co-ordinatesarea.
  *	w and h		Specifies width and height respectively.
  *
  * Returns:		none
  *
  * Comments:		none
  *
  * Sample application uses:
  *   - Window backgrounds.
  *   - x11perf -rect500 (rectangle tests).
  *   - x11perf -trap100 (fill trapezoid tests).
  *   - x11perf -hseg500 (horizontal line segments).
  *----------------------------------------------------------------------------
  */
void
McstMgaXAASubsequentSolidFillRect(ScrnInfoPtr pScrn, int x, int y,
                                   int width, int height)
{
    __SolidFill(pScrn, x, y, width, height);
}

void
McstMgaXAA_WritePixmap(ScrnInfoPtr pScrn, int x, int y, int w, int h, int rop,
                       unsigned int planemask, int transColor, int bpp,
                       int depth, int skip)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 ctrl;
    int cbppCd;                /* bits per pixel & byte swap */
    int wSize;
    int Bpp;                    /* bytes per pixel */

    Bpp = bpp >> 3;
    switch (Bpp) {
    case 4:
        cbppCd = BB_CTRL_BPP_CD_32;
        break;
    case 3:
        cbppCd = BB_CTRL_BPP_CD_24;
        break;
    case 2:
        cbppCd = BB_CTRL_BPP_CD_16;
        break;
    case 1:
        cbppCd = BB_CTRL_BPP_CD_8;
        break;
    default:
        return;
    };
    wSize = w * Bpp;
    ctrl = ROPMap[rop] | cbppCd | BB_CTRL_SDMA_EN | BB_CTRL_CMD_START;
    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, (h << 16) | wSize);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, infoPtr->DirectBuff.physAddr);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR,
                       (y * pScrn->displayWidth + x) * Bpp);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH,
                       wSize + skip + ((pScrn->displayWidth * Bpp) << 16));
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl);
}

/**
 * McstMgaXAAWritePixmap
 *
 * Description:	This routine is used to prepare image buffer for XAA write (Direct method)
 *
 * Parameters.
 *	pScrn	Screen handler pointer having screen information.
 *
 * Comments:		none
 *
 */
void
McstMgaXAAWritePixmap(ScrnInfoPtr pScrn, int x, int y, int w, int h,
                      unsigned char *src, int srcWidth, int rop,
                      unsigned int planemask, int trans, int bpp, int depth)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int skipLeft;
    int depthBpp;
    int cbppCd;
    int wSize;
    int size;
    int Bpp;
    CARD32 offSet;
    CARD32 pitch;
    CARD32 ctrl;
    CARD32 win;
    CARD32 sa;
    CARD32 da;

    Bpp = bpp >> 3;
    depthBpp = depth >> 3;
    skipLeft = (long) src & 0x03L;
#if USE_MCSTMGA_QUEUE
    if (skipLeft > 0) {
        if (Bpp == 3)
            skipLeft = 4 - skipLeft;
        else
            skipLeft /= Bpp;
        if (x >= skipLeft) {
            x -= skipLeft;
            w += skipLeft;
            if (Bpp == 3)
                src -= 3 * skipLeft;
            else
                src = (unsigned char *) ((long) src & ~0x03L);
        }
    };
#endif

    /* Копируем в DMA буфер из указаного буфера, возможно с пропусками (по виртуальным адресам) */
    wSize = w * Bpp;
    offSet = getDMAAddr(infoPtr, wSize * h);
    if ((srcWidth - wSize) > 0) {
        unsigned char *s = src;
        CARD32 d = offSet;
        int i = h;

        size = wSize * h;
        for (i = 0; i < h; i++) {
            memcpy(infoPtr->DirectBuff.virtAddr + d, s, wSize);
            s += srcWidth;
            d += wSize;
        }
    }
    else {
        size = srcWidth * h;
        memcpy(infoPtr->DirectBuff.virtAddr + offSet, src, size);
    }

    switch (depthBpp) {
    case 4:
        cbppCd = BB_CTRL_BPP_CD_32;
        break;
    case 3:
        cbppCd = BB_CTRL_BPP_CD_24;
        break;
    case 2:
        cbppCd = BB_CTRL_BPP_CD_16;
        break;
    case 1:
        cbppCd = BB_CTRL_BPP_CD_8;
        break;
    default:
        return;
    };
    ctrl = ROPMap[rop] | cbppCd | BB_CTRL_SDMA_EN | BB_CTRL_CMD_START;
    win = (h << 16) | wSize;
    sa = infoPtr->DirectBuff.physAddr + offSet;
    da = (y * pScrn->displayWidth + x) * Bpp;
    pitch = (wSize) | ((pScrn->displayWidth * Bpp) << 16);
    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    if (infoPtr->BusyDMAList[0].size == 0) {
        infoPtr->BusyDMAList[0].size = size;
        infoPtr->BusyDMAList[0].startAddr = offSet;
    }
    else {
        infoPtr->BusyDMAList[1].size = size;
        infoPtr->BusyDMAList[1].startAddr = offSet;
    };
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, win);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, sa);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, da);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, pitch);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl);
}

/**
 * McstMgaXAAReadPixmap
 *
 * Description:	This routine is used to prepare image buffer for XAA read (Direct method)
 *
 * Parameters.
 *	pScrn	Screen handler pointer having screen information.
 *
 * Comments:		none
 *
 */
void
McstMgaXAAReadPixmap(ScrnInfoPtr pScrn, int x, int y, int w, int h,
                     unsigned char *dest, int destWidth, int bpp, int depth)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int bytes_per_pixel;
    int cbppCd;
    int wSize;
    int size;
    unsigned char *src;
    CARD32 offSet;
    CARD32 pitch;
    CARD32 ctrl;
    CARD32 win;
    CARD32 sa;
    CARD32 da;
    int i;

    /* Копирование в DMA буффер (по физическому адресу) */
    bytes_per_pixel = bpp >> 3;
    wSize = w * bytes_per_pixel;
    size = wSize * h;
    win = (h << 16) | wSize;
    sa = (y * pScrn->displayWidth + x) * bytes_per_pixel;
    offSet = getDMAAddr(infoPtr, size);
    da = infoPtr->DirectBuff.physAddr + offSet;
    pitch = (wSize << 16) | (pScrn->displayWidth * bytes_per_pixel);
    switch (depth >> 3) {
    case 4:
        cbppCd = BB_CTRL_BPP_CD_32;
        break;
    case 3:
        cbppCd = BB_CTRL_BPP_CD_24;
        break;
    case 2:
        cbppCd = BB_CTRL_BPP_CD_16;
        break;
    case 1:
        cbppCd = BB_CTRL_BPP_CD_8;
        break;
    default:
        return;
    };
    ctrl = BB_CTRL_ROP_SRC | cbppCd | BB_CTRL_DDMA_EN | BB_CTRL_CMD_START;
    src = infoPtr->DirectBuff.virtAddr + offSet;

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    if (infoPtr->BusyDMAList[0].size == 0) {
        infoPtr->BusyDMAList[0].size = size;
        infoPtr->BusyDMAList[0].startAddr = offSet;
    }
    else {
        infoPtr->BusyDMAList[1].size = size;
        infoPtr->BusyDMAList[1].startAddr = offSet;
    }
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, win);
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, sa);
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, da);
    McstMgaMmio32Write(infoPtr, REG_BB_PITCH, pitch);
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl);

    /* Дожидаемся окончания DMA пересылок (записываем первый байт обратно в карту: */
    /* так будет гарантированно что всё считано из видеопамяти в память процессора) */
    ctrl = BB_CTRL_ROP_DST | cbppCd | BB_CTRL_SDMA_EN | BB_CTRL_CMD_START;
    McstMgaWaitForFIFO(infoPtr, 0);
    McstMgaMmio32Write(infoPtr, REG_BB_WINDOW, (1 << 16) | 1);  /* пересылка из одного байта */
    McstMgaMmio32Write(infoPtr, REG_BB_SADDR, (da));    /* da -> SADDR */
    McstMgaMmio32Write(infoPtr, REG_BB_DADDR, (sa));    /* sa -> DADDR */
    for (i = 0; i < 8; i++) {
        if (infoPtr->BusyDMAList[0].size == 0) {
            infoPtr->BusyDMAList[0].size = 1;
            infoPtr->BusyDMAList[0].startAddr = offSet;
        }
        else {
            infoPtr->BusyDMAList[1].size = 1;
            infoPtr->BusyDMAList[1].startAddr = offSet;
        };
        McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
        McstMgaMmio32Write(infoPtr, REG_BB_CTRL, ctrl);
    }
    McstMgaWaitForFIFO(infoPtr, 0);

    /* Копируем из DMA буфера в указаный буфер, возможно с пропусками (по виртуальным адресам) */
    if ((destWidth - wSize) > 0) {
        do {
            memcpy(dest, src, wSize);
            dest += destWidth;
            src += wSize;
        } while (--h > 0);
    }
    else {
        size = destWidth * h;
        memcpy(dest, src, size);
    }
    initDMAList(infoPtr);
}

/*  This requires all LEFT_EDGE clipping.  You get too many problems
    with reading past the edge of the pattern otherwise */

static void
WriteColumn(ScrnInfoPtr pScrn, unsigned char *pSrc, int x, int y, int w, int h,
            int xoff, int yoff, int pHeight, int srcWidth, int Bpp, int rop,
            unsigned int planemask, int trans, int bpp, int depth)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    unsigned char *dest;
    unsigned char *src;
    int wSize;
    int size;
    int skip;

    pSrc += (Bpp * xoff);
    src = pSrc + (yoff * srcWidth);

    wSize = w * Bpp;
    size = srcWidth * h;
    if ((srcWidth - wSize) > 0)
        skip = srcWidth - wSize;
    dest = infoPtr->DirectBuff.virtAddr;
    memcpy(dest, src, size);

    McstMgaXAA_WritePixmap(pScrn, x, y, w, h, rop, planemask, trans, bpp,
                           depth, skip);
}

void
McstMgaXAAFillImageWriteRects(ScrnInfoPtr pScrn, int rop,
                              unsigned int planemask, int nBox, BoxPtr pBox,
                              int xorg, int yorg, PixmapPtr pPix)
{
    int Bpp = pPix->drawable.bitsPerPixel >> 3;
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int pHeight = pPix->drawable.height;
    int pWidth = pPix->drawable.width;
    int srcWidth = pPix->devKind;
    XAAInfoRecPtr infoRec;
    int phaseY;
    int phaseX;
    int height;
    int width;
    int blitW;
    int x;

    ErrorF("McstMgaXAAFillImageWriteRects: nBox = %d\n", nBox);
    infoRec = infoPtr->pXAAInfo;
    while (nBox--) {
        x = pBox->x1;
        phaseY = (pBox->y1 - yorg) % pHeight;
        if (phaseY < 0)
            phaseY += pHeight;
        phaseX = (x - xorg) % pWidth;
        if (phaseX < 0)
            phaseX += pWidth;
        height = pBox->y2 - pBox->y1;
        width = pBox->x2 - x;

        while (1) {
            blitW = pWidth - phaseX;
            if (blitW > width)
                blitW = width;

            WriteColumn(pScrn, pPix->devPrivate.ptr, x, pBox->y1,
                        blitW, height, phaseX, phaseY, pHeight, srcWidth, Bpp,
                        rop, planemask, -1,
                        pPix->drawable.bitsPerPixel, pPix->drawable.depth);

            width -= blitW;
            if (!width)
                break;
            x += blitW;
            phaseX = (phaseX + blitW) % pWidth;
        };
        pBox++;
    };

    McstMgaWaitForFIFO(infoPtr, USE_MCSTMGA_QUEUE);
    if (infoRec->ImageWriteFlags & SYNC_AFTER_IMAGE_WRITE)
        (*infoRec->Sync) (pScrn);
    else
        SET_SYNC_FLAG(infoRec);
}

/**
 * McstMgaXAAWriteImageDirectInit
 *
 * Description:	This routine is used to initialize XAA write image Direct method
 *
 * Parameters.
 *	pScreenInfo	Screen handler pointer having screen information.
 *
 * Returns:		-1 - error. Can not allocate dma buffers.
 *			0  - ok
 * Comments:		none
 *
 */
static int
McstMgaXAAWriteImageDirectInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    /* Need DMA support */

    ErrorF(" ################# new #####################\n");
    infoPtr->pXAAInfo->WritePixmap = McstMgaXAAWritePixmap;
    infoPtr->pXAAInfo->ReadPixmap = McstMgaXAAReadPixmap;

    return (0);
}

/**
 * McstMgaXAAWriteImageDirectInit
 *
 * Description:	This routine is used to initialize XAA "Indirect CPU to
 *              Screen color expansion" method
 *
 * Parameters.
 *	pScreenInfo	Screen handler pointer having screen information.
 *
 * Returns:		-1 - error. Can not allocate dma buffers.
 *			0  - ok
 * Comments:		none
 *
 */
static int
McstMgaXAAIndirectCPU2ScreenColorExpansionInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    /* The Indirect CPU To Screen Color Expansion (use DMA from system memory) */
    infoPtr->pXAAInfo->NumScanlineColorExpandBuffers = 4;
    _scanline_buf[0] =
        infoPtr->DirectBuff.virtAddr + MCSTMGA_MEM_SIZE +
        (0 * MCSTMGA_SCANLINE_BUF_SIZE);
    _scanline_buf[1] =
        infoPtr->DirectBuff.virtAddr + MCSTMGA_MEM_SIZE +
        (1 * MCSTMGA_SCANLINE_BUF_SIZE);
    _scanline_buf[2] =
        infoPtr->DirectBuff.virtAddr + MCSTMGA_MEM_SIZE +
        (2 * MCSTMGA_SCANLINE_BUF_SIZE);
    _scanline_buf[3] =
        infoPtr->DirectBuff.virtAddr + MCSTMGA_MEM_SIZE +
        (3 * MCSTMGA_SCANLINE_BUF_SIZE);
    infoPtr->pXAAInfo->ScanlineColorExpandBuffers = _scanline_buf;
    infoPtr->pXAAInfo->SetupForScanlineCPUToScreenColorExpandFill =
        McstMgaXAASetupForScanlineCPUToScreenColorExpandFill;
    infoPtr->pXAAInfo->SubsequentScanlineCPUToScreenColorExpandFill =
        McstMgaXAASubsequentScanlineCPUToScreenColorExpandFill;
    infoPtr->pXAAInfo->SubsequentColorExpandScanline =
        McstMgaXAASubsequentColorExpandScanlinee;
    infoPtr->pXAAInfo->ScanlineCPUToScreenColorExpandFillFlags = 0
        | BIT_ORDER_IN_BYTE_LSBFIRST | SCANLINE_PAD_DWORD
        ;

    return (0);
}

/**
 * XAA Acceleration init
 * @param pScreenInfo	Screeen info pointer structure.
 * @return		success flag
 */
static Bool
McstMgaXAA_Init(ScreenPtr pScreen, int xaaFlags)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int fd = fbdevHWGetFD(pScrn);
    struct dmaMem dmem;

    /* Creates XAA record */
    if (!(infoPtr->pXAAInfo = XAACreateInfoRec())) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "XAACreateInfoRec Error\n");
        return FALSE;
    }

    /* Flags */
    /* accelerator does not support skipping colors when copying */
    infoPtr->pXAAInfo->Flags = xaaFlags | LINEAR_FRAMEBUFFER | NO_PLANEMASK;

    infoPtr->pXAAInfo->PixmapCacheFlags |= DO_NOT_BLIT_STIPPLES;
    /* Sync */
    infoPtr->pXAAInfo->Sync = McstMgaWaitForIdle;

    /* Screen-to-screen copy */
    infoPtr->pXAAInfo->SetupForScreenToScreenCopy =
        McstMgaXAASetupForScreenToScreenCopy;
    infoPtr->pXAAInfo->SubsequentScreenToScreenCopy =
        McstMgaXAASubsequentScreenToScreenCopy;
    infoPtr->pXAAInfo->ScreenToScreenCopyFlags = NO_TRANSPARENCY;

    /* Screen-to-screen color expand fill */
    infoPtr->pXAAInfo->SetupForScreenToScreenColorExpandFill =
        McstMgaXAASetupForScreenToScreenColorExpandFill;
    infoPtr->pXAAInfo->SubsequentScreenToScreenColorExpandFill =
        McstMgaXAASubsequentScreenToScreenColorExpandFill;
    infoPtr->pXAAInfo->ScreenToScreenColorExpandFillFlags = 0;

    /* Solid fills */
    /* XAA uses solid fill for horizontal and vertical solid lines.
     * Bresenham lines are not supported by hardware
     */
    infoPtr->pXAAInfo->SetupForSolidFill = McstMgaXAASetupForSolidFill;
    infoPtr->pXAAInfo->SubsequentSolidFillRect =
        McstMgaXAASubsequentSolidFillRect;
    infoPtr->pXAAInfo->SolidFillFlags = 0;
    if (infoPtr->NoDMA == FALSE) {
#ifdef USE_XAA_DMA
        if (XAA_GetDMAMemory(pScreen) == 0) {
            McstMgaXAAWriteImageDirectInit(pScreen);
            McstMgaXAAIndirectCPU2ScreenColorExpansionInit(pScreen);
        }
#endif
    };

    if (!XAAInit(pScreen, infoPtr->pXAAInfo)) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "XAAInit Error\n");
        return FALSE;
    }
    return TRUE;
}

/**
 * XAA Acceleration cleanup
 * @param pScreenInfo	Screeen info pointer structure.
 * @return		none
 */
static void
McstMgaXAA_Cleanup(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    XAADestroyInfoRec(infoPtr->pXAAInfo);
    infoPtr->pXAAInfo = NULL;
}
#endif /* USE_XAA */

/*******************************************************************************
 * Acceleration init & cleanup
 ******************************************************************************/

/**
 * MCST MGA acceleration initialization
 */
Bool
McstMgaAccelInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 r;

    /* Firmware version 0 hangs videocard bitblt module when using it */
    if (infoPtr->firmwareVersion == 0) {
        xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                   "Firmware version 0 does not support acceleration\n");
        return FALSE;
    }

#ifdef MCSTMGA_DEBUG_ACCEL
    memset(infoPtr->dumpBlock, 0, sizeof(infoPtr->dumpBlock));
    infoPtr->dumpBlockIdx = 0;
#endif

    /* Reseting bitblt module */
    McstMgaMmio32Write(infoPtr, REG_BB_CTRL, BB_CTRL_CMD_ABORT);
    infoPtr->MGA_regs.wind = -1;

    r = McstMgaMmio32Read(infoPtr, REG_BB_STAT);
    if ((r & BB_STAT_DMA) == 0) {
        xf86DrvMsg(pScrn->scrnIndex, X_INFO,
                   "DMA disabled (this hardware doesn't support it)\n");
        infoPtr->NoDMA = TRUE;
    }
    if (infoPtr->useEXA) {
        if (!McstMgaEXAInit(pScreen))
            return FALSE;
    };
#ifdef USE_XAA
    if (infoPtr->useXAA) {
        int xaaFlags = 0;
        BoxRec box;

        box.x1 = 0;
        box.y1 = 0;
        box.x2 = pScrn->displayWidth;
        box.y2 = (infoPtr->fbmem_len - MCSTMGA_CURSOR_SIZE)
            / (pScrn->displayWidth * pScrn->bitsPerPixel >> 3);

        if (xf86InitFBManager(pScreen, &box)) {
            xaaFlags |= PIXMAP_CACHE | OFFSCREEN_PIXMAPS;
            /*infoPtr->pXAAInfo->CachePixelGranularity = 4; */
        }
        else {
            xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
                       "Failed to init memory manager\n");
        }

        if (!McstMgaXAA_Init(pScreen, xaaFlags))
            return FALSE;
    }
#endif /* USE_XAA */

    return TRUE;
}

/**
 * MCST MGA acceleration cleanup
 */
void
McstMgaAccelCleanup(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int fd = fbdevHWGetFD(pScrn);

    if (infoPtr->useEXA) {
        McstMgaEXA_Cleanup(pScreen);
        munmap(infoPtr->DmaChunkVBase, infoPtr->DmaChunkRange);
    }
#ifdef USE_XAA
    if (infoPtr->useXAA) {
        McstMgaXAA_Cleanup(pScreen);
    }
#endif /* USE_XAA */
}
