/********************************************
*
* Copyright (c) 2022-2024 AO MCST
* Distributed under the terms of MIT License.
*
*********************************************/

#include <xorg-server.h>
#include "xf86.h"
#include "xf86_OSproc.h"

#include "mcstmga.h"
#include "cursorstr.h"

// #define MCSTMGA_DEBUG_CURSOR

static void
McstMgaShowCursor(ScrnInfoPtr pScrn)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 ctrl = McstMgaMmio32Read(infoPtr, REG_CTRL);

    ctrl |= CTRL_HC0E;
    McstMgaMmio32Write(infoPtr, REG_CTRL, ctrl);
}

static void
McstMgaHideCursor(ScrnInfoPtr pScrn)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    CARD32 ctrl = McstMgaMmio32Read(infoPtr, REG_CTRL);

    ctrl &= ~(CARD32) CTRL_HC0E;
    McstMgaMmio32Write(infoPtr, REG_CTRL, ctrl);
}

static void
McstMgaSetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    McstMgaMmio32Write(infoPtr, REG_C0XY, (y & 0xffff) << 16 | (x & 0xffff));
}

/* Firmware version 0 does not support negative cursor coordinates */
static void
McstMgaSetCursorPosition_fw_0(ScrnInfoPtr pScrn, int x, int y)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    x = (x < 0) ? 0 : x;
    y = (y < 0) ? 0 : y;
    McstMgaMmio32Write(infoPtr, REG_C0XY, (y & 0xffff) << 16 | (x & 0xffff));
}

static void
McstMgaSetCursorColors(ScrnInfoPtr pScrn, int bg, int fg)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

#ifdef MCSTMGA_DEBUG_CURSOR
    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "cursor: bg: 0x%08lx fg: 0x%08lx\n",
               bg, fg);
#endif

#define PACK_COLOR(rgb) \
    ((rgb & 0x00f80000) >> 9 | (rgb & 0x0000f800) >> 6 | (rgb & 0x000000f8) >> 3)

    McstMgaMmio32Write(infoPtr, REG_C0CR + 0x04, PACK_COLOR(fg));
    McstMgaMmio32Write(infoPtr, REG_C0CR + 0x00, PACK_COLOR(bg) << 16 | 0x8000);

#define REG_Bug48595 0xf8
    McstMgaMmio32Write(infoPtr, REG_Bug48595, 0x0);
#undef REG_Bug48595

#undef PACK_COLOR
}

static void
McstMgaLoadCursorImage(ScrnInfoPtr pScrn, unsigned char *bits)
{
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);
    int iColor = 0x0;    /* Target pixel color index */
    int iSrc;                   /* Source pixel index [2 bits size] */
    int iDst;                   /* Target pixel index [4 bits size] */
    CARD32 packedDst = 0;
    CARD32 ctrl;
    CARD32 res;
    volatile CARD32 *dest =
        (volatile CARD32 *) (infoPtr->fbmem + infoPtr->CursorImageOffset);
#ifdef MCSTMGA_DEBUG_CURSOR
    char dbg[65] = { 0, };
#endif

    ctrl = McstMgaMmio32Read(infoPtr, REG_CTRL);
    for (iSrc = 0, iDst = 0; iDst < 0x1000; iSrc++, iDst++) {
        switch (((bits[iSrc >> 2]) >> ((3 - iSrc % 4) * 2)) & 0x03) {
        case 0x00:             /* pass trough */
        case 0x01:
            iColor = 0x00;
            break;
        case 0x02:
            iColor = 0x01;
            break;
        case 0x03:
            iColor = 0x02;
            break;
        }

#ifdef __sparc__
        packedDst |= iColor << (4 * (iDst % 8));
#else
        packedDst |= iColor << (4 * (7 - iDst % 8));
#endif

        if (iDst != 0 && iDst % 8 == 7) {
            switch (pScrn->bitsPerPixel) {
            case 8:
            case 24:
#ifdef __sparc__
                *dest = McstMgaFlip32(packedDst);
#else
                *dest = packedDst;
#endif
                break;
            case 16:
#ifdef __sparc__
                *dest = packedDst;
#else
                res = ((packedDst & 0xFFFF) << 16) |
                    ((packedDst >> 16) & 0xFFFF);
                *dest = res;
#endif
                break;
            case 32:
            default:
                *dest = packedDst;
                break;
            };
            dest++;
            packedDst = 0;
        }

#ifdef MCSTMGA_DEBUG_CURSOR
        switch (iColor) {
        case 0x01:
            dbg[iDst % 64] = '.';
            break;
        case 0x02:
            dbg[iDst % 64] = '#';
            break;
        default:
            dbg[iDst % 64] = ' ';
            break;
        }

        if (iDst != 0 && iDst % 64 == 0) {
#ifdef __E2K__
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[7 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[15 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[23 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[31 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[39 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[47 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[55 - i]);
            for (i = 0; i < 8; i++)
                ErrorF("%c", dbg[63 - i]);
            ErrorF("\n");
#else
            xf86Msg(X_ERROR, "%s\n", dbg);
#endif
        }
#endif
    }

    ctrl |= CTRL_HC0R_64;
    McstMgaMmio32Write(infoPtr, REG_CTRL, ctrl);
    McstMgaMmio32Write(infoPtr, REG_C0BAR, infoPtr->CursorImageOffset);
}

Bool
McstMgaHWCursorInit(ScreenPtr pScreen)
{
    xf86CursorInfoPtr cursorPtr = xf86CreateCursorInfoRec();
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    McstMgaPtr infoPtr = McstMgaGetRec(pScrn);

    if (!cursorPtr)
        return FALSE;

    cursorPtr->MaxWidth = 64;
    cursorPtr->MaxHeight = 64;
    cursorPtr->SetCursorColors = McstMgaSetCursorColors;

    /* Firmware version 0 does not support negative cursor coords */
    if (infoPtr->firmwareVersion == 0) {
        cursorPtr->SetCursorPosition = McstMgaSetCursorPosition_fw_0;
    }
    else {
        cursorPtr->SetCursorPosition = McstMgaSetCursorPosition;
    }

    cursorPtr->LoadCursorImage = McstMgaLoadCursorImage;
    cursorPtr->HideCursor = McstMgaHideCursor;
    cursorPtr->ShowCursor = McstMgaShowCursor;

    cursorPtr->Flags = HARDWARE_CURSOR_AND_SOURCE_WITH_MASK
        | HARDWARE_CURSOR_NIBBLE_SWAPPED
        | HARDWARE_CURSOR_TRUECOLOR_AT_8BPP
        | HARDWARE_CURSOR_SOURCE_MASK_INTERLEAVE_1;

    if (!xf86InitCursor(pScreen, cursorPtr))
        return FALSE;

    infoPtr->CursorInfoRec = cursorPtr;
    infoPtr->CursorImageOffset = infoPtr->fbmem_len - MCSTMGA_CURSOR_SIZE;

    return TRUE;
}
